import logging

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Count, signals
from django.db import transaction
from django.utils.timezone import now

from waldur_core.core import utils as core_utils
from waldur_core.structure.models import Project, Customer

from . import callbacks, tasks, log, models, utils


logger = logging.getLogger(__name__)


def create_screenshot_thumbnail(sender, instance, created=False, **kwargs):
    if not created:
        return

    transaction.on_commit(lambda: tasks.create_screenshot_thumbnail.delay(instance.uuid))


def log_order_events(sender, instance, created=False, **kwargs):
    order = instance
    if created:
        # Skip logging for imported orders
        if order.state != models.Order.States.DONE:
            log.log_order_created(order)
    elif not order.tracker.has_changed('state'):
        return
    elif order.state == models.Order.States.EXECUTING:
        log.log_order_approved(order)
    elif order.state == models.Order.States.REJECTED:
        log.log_order_rejected(order)
    elif order.state == models.Order.States.DONE:
        log.log_order_completed(order)
    elif order.state == models.Order.States.TERMINATED:
        log.log_order_terminated(order)
    elif order.state == models.Order.States.ERRED:
        log.log_order_failed(order)


def log_order_item_events(sender, instance, created=False, **kwargs):
    order_item = instance
    if not created:
        return
    if order_item.state != models.OrderItem.States.PENDING:
        # Skip logging for imported order item
        return
    elif not order_item.resource:
        return
    elif order_item.type == models.OrderItem.Types.TERMINATE:
        log.log_resource_terminate_requested(order_item.resource)
    elif order_item.type == models.OrderItem.Types.UPDATE:
        log.log_resource_update_requested(order_item.resource)


def log_resource_events(sender, instance, created=False, **kwargs):
    resource = instance
    # Skip logging for imported resource
    if created and instance.state == models.Resource.States.CREATING:
        log.log_resource_creation_requested(resource)


def complete_order_when_all_items_are_done(sender, instance, created=False, **kwargs):
    if created:
        return

    if not instance.tracker.has_changed('state'):
        return

    if instance.state not in models.OrderItem.States.TERMINAL_STATES:
        return

    if instance.order.state != models.Order.States.EXECUTING:
        return

    order = instance.order
    # check if there are any non-finished OrderItems left and finish order if none is found
    if models.OrderItem.objects.filter(order=order).\
            exclude(state__in=models.OrderItem.States.TERMINAL_STATES).exists():
        return

    order.complete()
    order.save(update_fields=['state'])


def update_category_quota_when_offering_is_created(sender, instance, created=False, **kwargs):
    def get_delta():
        if created:
            if instance.state == models.Offering.States.ACTIVE:
                return 1
        else:
            if instance.tracker.has_changed('state'):
                if instance.state == models.Offering.States.ACTIVE:
                    return 1
                elif instance.tracker.previous('state') == models.Offering.States.ACTIVE:
                    return -1

    delta = get_delta()
    if delta:
        instance.category.add_quota_usage(models.Category.Quotas.offering_count, delta)


def update_category_quota_when_offering_is_deleted(sender, instance, **kwargs):
    if instance.state == models.Offering.States.ACTIVE:
        instance.category.add_quota_usage(models.Category.Quotas.offering_count, -1)


def update_category_offerings_count(sender, **kwargs):
    for category in models.Category.objects.all():
        value = models.Offering.objects.filter(category=category,
                                               state=models.Offering.States.ACTIVE).count()
        category.set_quota_usage(models.Category.Quotas.offering_count, value)


def update_aggregate_resources_count_when_resource_is_updated(sender, instance, created=False, **kwargs):
    def apply_change(delta):
        for field in ('project', 'customer'):
            try:
                scope = getattr(instance, field)
            except ObjectDoesNotExist:
                # When project is deleted, all its resources are deleted via cascade.
                # Therefore it is okay if project does not exists.
                continue
            counter, _ = models.AggregateResourceCount.objects.get_or_create(
                scope=scope,
                category=instance.offering.category,
            )
            if delta == 1:
                counter.count += 1
            elif delta == -1:
                counter.count = max(0, counter.count - 1)

            counter.save(update_fields=['count'])

    if created and instance.state != models.Resource.States.TERMINATED:
        apply_change(1)
    elif instance.tracker.has_changed('state') and instance.state == models.Resource.States.TERMINATED:
        apply_change(-1)


def update_aggregate_resources_count(sender, **kwargs):
    for category in models.Category.objects.all():
        for field, content_type in (
            ('project_id', ContentType.objects.get_for_model(Project)),
            ('project__customer_id', ContentType.objects.get_for_model(Customer)),
        ):
            rows = models.Resource.objects\
                .filter(offering__category=category)\
                .exclude(state=models.Resource.States.TERMINATED)\
                .values(field, 'offering__category')\
                .annotate(count=Count('*'))
            for row in rows:
                models.AggregateResourceCount.objects.update_or_create(
                    content_type=content_type,
                    object_id=row[field],
                    category=category,
                    defaults={'count': row['count']},
                )


def close_resource_plan_period_when_resource_is_terminated(sender, instance, created=False, **kwargs):
    """
    Handle case when resource has been terminated by service provider.
    """

    if created:
        return

    if not instance.tracker.has_changed('state'):
        return

    if instance.state != models.Resource.States.TERMINATED:
        return

    if instance.tracker.previous('state') == models.Resource.States.TERMINATING:
        # It is expected that this case is handled using callbacks
        return

    if not instance.plan:
        return

    models.ResourcePlanPeriod.objects.filter(
        resource=instance,
        plan=instance.plan,
        end=None
    ).update(end=now())


def reject_order(sender, instance, created=False, **kwargs):
    if created:
        return

    order = instance

    if not order.tracker.has_changed('state'):
        return

    if instance.tracker.previous('state') == models.Order.States.REQUESTED_FOR_APPROVAL and \
            order.state == models.Order.States.REJECTED:
        for item in order.items.all():
            item.set_state_terminated()
            item.save(update_fields=['state'])


def change_order_item_state(sender, instance, created=False, **kwargs):
    if created or not instance.tracker.has_changed('state'):
        return

    try:
        resource = models.Resource.objects.get(scope=instance)
    except ObjectDoesNotExist:
        logger.warning('Skipping resource state synchronization '
                       'because marketplace resource is not found. '
                       'Resource ID: %s', core_utils.serialize_instance(instance))
    else:
        callbacks.sync_resource_state(instance, resource)


def terminate_resource(sender, instance, **kwargs):
    try:
        resource = models.Resource.objects.get(scope=instance)
    except ObjectDoesNotExist:
        logger.debug('Skipping terminate for resource '
                     'because marketplace resource does not exist. '
                     'Resource ID: %s', core_utils.serialize_instance(instance))
    else:
        callbacks.resource_deletion_succeeded(resource)


def connect_resource_handlers(*resources):
    for index, model in enumerate(resources):
        suffix = '%s_%s' % (index, model.__class__)

        signals.post_save.connect(
            change_order_item_state,
            sender=model,
            dispatch_uid='waldur_mastermind.marketpace.change_order_item_state_%s' % suffix,
        )

        signals.pre_delete.connect(
            terminate_resource,
            sender=model,
            dispatch_uid='waldur_mastermind.marketpace.terminate_resource_%s' % suffix,
        )


def synchronize_resource_metadata(sender, instance, created=False, **kwargs):
    fields = {'action', 'action_details', 'state', 'runtime_state', 'name'}
    if not created and not set(instance.tracker.changed()) & fields:
        return

    try:
        resource = models.Resource.objects.get(scope=instance)
    except ObjectDoesNotExist:
        logger.debug('Skipping resource synchronization for OpenStack resource '
                     'because marketplace resource does not exist. '
                     'Resource ID: %s', instance.id)
        return

    utils.import_resource_metadata(resource)


def connect_resource_metadata_handlers(*resources):
    for index, model in enumerate(resources):
        signals.post_save.connect(
            synchronize_resource_metadata,
            sender=model,
            dispatch_uid='waldur_mastermind.marketpace.'
                         'synchronize_resource_metadata_%s_%s' % (index, model.__class__),
        )


@transaction.atomic()
def limit_update_succeeded(sender, order_item, **kwargs):
    resource = order_item.resource
    old_limits = resource.limits
    resource.limits = order_item.limits
    resource.save()
    order_item.set_state_done()
    order_item.save(update_fields=['state'])
    logger.info('Resource limits have been updated. Resource: %s, old limits: %s, new limits: %s, created by: %s',
                core_utils.serialize_instance(resource),
                old_limits,
                resource.limits,
                order_item.order.created_by)
    log.log_resource_limit_update_succeeded(resource)


def limit_update_failed(sender, order_item, error_message, **kwargs):
    order_item.set_state_erred()
    order_item.error_message = error_message
    order_item.save()
    resource = order_item.resource
    logger.info('Resource limit update failed. Resource: %s, requested limits: %s, created by: %s, '
                'error message: %s',
                core_utils.serialize_instance(resource),
                resource.limits,
                order_item.order.created_by,
                error_message)
    log.log_resource_limit_update_failed(resource)



from datetime import datetime
import logging
import re

from croniter.croniter import croniter
from cryptography.exceptions import UnsupportedAlgorithm
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import serialization
from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import PermissionsMixin, UserManager
from django.contrib.postgres.fields import JSONField as BetterJSONField
from django.core import validators
from django.core.exceptions import ValidationError
from django.core.mail import send_mail
from django.db import models, transaction
from django.utils import timezone as django_timezone
from django.utils.encoding import force_text
from django.utils.lru_cache import lru_cache
from django.utils.translation import ugettext_lazy as _
from django_fsm import transition, FSMIntegerField
from model_utils import FieldTracker
from model_utils.models import TimeStampedModel
import pytz
from reversion import revisions as reversion
from reversion.models import Version

from waldur_core.core.fields import CronScheduleField, UUIDField
from waldur_core.core.validators import validate_name, MinCronValueValidator
from waldur_core.logging.loggers import LoggableMixin

from .shims import AbstractBaseUser

logger = logging.getLogger(__name__)


class DescribableMixin(models.Model):
    """
    Mixin to add a standardized "description" field.
    """
    class Meta:
        abstract = True

    description = models.CharField(_('description'), max_length=500, blank=True)


class NameMixin(models.Model):
    """
    Mixin to add a standardized "name" field.
    """

    class Meta:
        abstract = True

    name = models.CharField(_('name'), max_length=150, validators=[validate_name])


class UiDescribableMixin(DescribableMixin):
    """
    Mixin to add a standardized "description" and "icon url" fields.
    """
    class Meta:
        abstract = True

    icon_url = models.URLField(_('icon url'), blank=True)


class UuidMixin(models.Model):
    """
    Mixin to identify models by UUID.
    """
    class Meta:
        abstract = True

    uuid = UUIDField()


class ErrorMessageMixin(models.Model):
    """
    Mixin to add a standardized "error_message" field.
    """
    class Meta:
        abstract = True

    error_message = models.TextField(blank=True)


class ScheduleMixin(models.Model):
    """
    Mixin to add a standardized "schedule" fields.
    """
    class Meta:
        abstract = True

    schedule = CronScheduleField(max_length=15, validators=[MinCronValueValidator(1)])
    next_trigger_at = models.DateTimeField(null=True)
    timezone = models.CharField(max_length=50, default=django_timezone.get_current_timezone_name)
    is_active = models.BooleanField(default=False)

    def update_next_trigger_at(self):
        tz = pytz.timezone(self.timezone)
        dt = tz.normalize(django_timezone.now())
        self.next_trigger_at = croniter(self.schedule, dt).get_next(datetime)

    def save(self, *args, **kwargs):
        """
        Updates next_trigger_at field if:
         - instance become active
         - instance.schedule changed
         - instance is new
        """
        try:
            prev_instance = self.__class__.objects.get(pk=self.pk)
        except self.DoesNotExist:
            prev_instance = None

        if prev_instance is None or (not prev_instance.is_active and self.is_active or
                                     self.schedule != prev_instance.schedule or
                                     self.timezone != prev_instance.timezone):
            self.update_next_trigger_at()

        super(ScheduleMixin, self).save(*args, **kwargs)


class UserDetailsMixin(models.Model):
    """
    This mixin is shared by User and Invitation model. All fields are optional.
    User is populated with these details when invitation is approved.
    Note that civil_number and email fields are not included in this mixin
    because they have different constraints in User and Invitation model.
    """
    class Meta:
        abstract = True

    full_name = models.CharField(_('full name'), max_length=100, blank=True)
    native_name = models.CharField(_('native name'), max_length=100, blank=True)
    phone_number = models.CharField(_('phone number'), max_length=255, blank=True)
    organization = models.CharField(_('organization'), max_length=255, blank=True)
    job_title = models.CharField(_('job title'), max_length=40, blank=True)


class User(LoggableMixin, UuidMixin, DescribableMixin, AbstractBaseUser, UserDetailsMixin, PermissionsMixin):
    username = models.CharField(
        _('username'), max_length=128, unique=True,
        help_text=_('Required. 128 characters or fewer. Letters, numbers and '
                    '@/./+/-/_ characters'),
        validators=[
            validators.RegexValidator(re.compile(r'^[\w.@+-]+$'), _('Enter a valid username.'), 'invalid')
        ])
    # Civil number is nullable on purpose, otherwise
    # it wouldn't be possible to put a unique constraint on it
    civil_number = models.CharField(_('civil number'), max_length=50, unique=True, blank=True, null=True, default=None)
    email = models.EmailField(_('email address'), max_length=75, blank=True)

    is_staff = models.BooleanField(_('staff status'), default=False,
                                   help_text=_('Designates whether the user can log into this admin '
                                               'site.'))
    is_active = models.BooleanField(_('active'), default=True,
                                    help_text=_('Designates whether this user should be treated as '
                                                'active. Unselect this instead of deleting accounts.'))
    is_support = models.BooleanField(_('support status'), default=False,
                                     help_text=_('Designates whether the user is a global support user.'))
    date_joined = models.DateTimeField(_('date joined'), default=django_timezone.now)
    registration_method = models.CharField(_('registration method'), max_length=50, default='default', blank=True,
                                           help_text=_('Indicates what registration method were used.'))
    agreement_date = models.DateTimeField(_('agreement date'), blank=True, null=True,
                                          help_text=_('Indicates when the user has agreed with the policy.'))
    preferred_language = models.CharField(max_length=10, blank=True)
    competence = models.CharField(max_length=255, blank=True)
    token_lifetime = models.PositiveIntegerField(null=True, help_text=_('Token lifetime in seconds.'),
                                                 validators=[validators.MinValueValidator(60)])
    details = BetterJSONField(blank=True,
                              default=dict,
                              help_text=_('Extra details from authentication backend.'))
    backend_id = models.CharField(max_length=255, blank=True)

    tracker = FieldTracker()
    objects = UserManager()

    USERNAME_FIELD = 'username'
    REQUIRED_FIELDS = ['email']

    class Meta:
        verbose_name = _('user')
        verbose_name_plural = _('users')

    def get_log_fields(self):
        return ('uuid', 'full_name', 'native_name', self.USERNAME_FIELD, 'is_staff', 'is_support', 'token_lifetime')

    def get_full_name(self):
        # This method is used in django-reversion as name of revision creator.
        return self.full_name

    def get_short_name(self):
        # This method is used in django-reversion as name of revision creator.
        return self.full_name

    def email_user(self, subject, message, from_email=None):
        """
        Sends an email to this User.
        """
        send_mail(subject, message, from_email, [self.email])

    @classmethod
    def get_permitted_objects(cls, user):
        from waldur_core.structure.filters import filter_visible_users

        queryset = User.objects.all()
        if user.is_staff or user.is_support:
            return queryset
        else:
            return filter_visible_users(queryset, user)

    def clean(self):
        super(User, self).clean()
        # User email has to be unique or empty
        if self.email and User.objects.filter(email=self.email).exclude(id=self.id).exists():
            raise ValidationError({'email': _('User with email "%s" already exists.') % self.email})

    @transaction.atomic
    def create_request_for_update_email(self, email):
        if User.objects.filter(email=email).exclude(id=self.id).exists():
            raise ValidationError({'email': _('User with email "%s" already exists.') % email})

        ChangeEmailRequest.objects.filter(user=self).delete()
        change_request = ChangeEmailRequest.objects.create(
            user=self,
            email=email,
        )
        return change_request

    def __str__(self):
        if self.full_name:
            return '%s (%s)' % (self.get_username(), self.full_name)

        return self.get_username()


class ChangeEmailRequest(UuidMixin, TimeStampedModel):
    user = models.OneToOneField(User, on_delete=models.CASCADE)
    email = models.EmailField()

    class Meta:
        verbose_name = _('change email request')
        verbose_name_plural = _('change email requests')


def validate_ssh_public_key(ssh_key):
    if isinstance(ssh_key, str):
        ssh_key = ssh_key.encode('utf-8')

    try:
        serialization.load_ssh_public_key(ssh_key, backends.default_backend())
    except (ValueError, UnsupportedAlgorithm) as e:
        logger.debug('Invalid SSH public key %s. Error: %s', ssh_key, e)
        raise ValidationError(_('Invalid SSH public key.'))


def get_ssh_key_fingerprint(ssh_key):
    # How to get fingerprint from ssh key:
    # http://stackoverflow.com/a/6682934/175349
    # http://www.ietf.org/rfc/rfc4716.txt Section 4.
    import base64
    import hashlib

    key_body = base64.b64decode(ssh_key.strip().split()[1].encode('ascii'))
    fp_plain = hashlib.md5(key_body).hexdigest()  # nosec
    return ':'.join(a + b for a, b in zip(fp_plain[::2], fp_plain[1::2]))


class SshPublicKey(LoggableMixin, UuidMixin, models.Model):
    """
    User public key.

    Used for injection into VMs for remote access.
    """
    user = models.ForeignKey(on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL, db_index=True)
    # Model doesn't inherit NameMixin, because name field can be blank.
    name = models.CharField(max_length=150, blank=True)
    fingerprint = models.CharField(max_length=47)  # In ideal world should be unique
    public_key = models.TextField(
        validators=[validators.MaxLengthValidator(2000), validate_ssh_public_key]
    )
    is_shared = models.BooleanField(default=False)

    class Meta:
        unique_together = ('user', 'name')
        verbose_name = _('SSH public key')
        verbose_name_plural = _('SSH public keys')

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        # Fingerprint is always set based on public_key
        try:
            self.fingerprint = get_ssh_key_fingerprint(self.public_key)
        except (IndexError, TypeError):
            logger.exception('Fingerprint calculation has failed')
            raise ValueError(_('Public key format is incorrect. Fingerprint calculation has failed.'))

        if update_fields and 'public_key' in update_fields and 'fingerprint' not in update_fields:
            update_fields.append('fingerprint')

        super(SshPublicKey, self).save(force_insert, force_update, using, update_fields)

    def __str__(self):
        return '%s - %s, user: %s, %s' % (self.name, self.fingerprint, self.user.username, self.user.full_name)


class RuntimeStateMixin(models.Model):
    """ Provide runtime_state field """
    class RuntimeStates:
        ONLINE = 'online'
        OFFLINE = 'offline'

    class Meta:
        abstract = True

    runtime_state = models.CharField(_('runtime state'), max_length=150, blank=True)

    @classmethod
    def get_online_state(cls):
        return cls.RuntimeStates.ONLINE

    @classmethod
    def get_offline_state(cls):
        return cls.RuntimeStates.OFFLINE


class StateMixin(ErrorMessageMixin):
    class States:
        CREATION_SCHEDULED = 5
        CREATING = 6
        UPDATE_SCHEDULED = 1
        UPDATING = 2
        DELETION_SCHEDULED = 7
        DELETING = 8
        OK = 3
        ERRED = 4

        CHOICES = (
            (CREATION_SCHEDULED, 'Creation Scheduled'),
            (CREATING, 'Creating'),
            (UPDATE_SCHEDULED, 'Update Scheduled'),
            (UPDATING, 'Updating'),
            (DELETION_SCHEDULED, 'Deletion Scheduled'),
            (DELETING, 'Deleting'),
            (OK, 'OK'),
            (ERRED, 'Erred'),
        )

    class Meta:
        abstract = True

    state = FSMIntegerField(
        default=States.CREATION_SCHEDULED,
        choices=States.CHOICES,
    )

    @property
    def human_readable_state(self):
        return force_text(dict(self.States.CHOICES)[self.state])

    @transition(field=state, source=States.CREATION_SCHEDULED, target=States.CREATING)
    def begin_creating(self):
        pass

    @transition(field=state, source=States.UPDATE_SCHEDULED, target=States.UPDATING)
    def begin_updating(self):
        pass

    @transition(field=state, source=States.DELETION_SCHEDULED, target=States.DELETING)
    def begin_deleting(self):
        pass

    @transition(field=state, source=[States.OK, States.ERRED], target=States.UPDATE_SCHEDULED)
    def schedule_updating(self):
        pass

    @transition(field=state, source=[States.OK, States.ERRED], target=States.DELETION_SCHEDULED)
    def schedule_deleting(self):
        pass

    @transition(field=state, source=[States.UPDATING, States.CREATING],
                target=States.OK)
    def set_ok(self):
        pass

    @transition(field=state, source='*', target=States.ERRED)
    def set_erred(self):
        pass

    @transition(field=state, source=States.ERRED, target=States.OK)
    def recover(self):
        pass

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]


class ReversionMixin:
    """ Store historical values of instance, using django-reversion.

        Note: `ReversionMixin` model should be registered in django-reversion,
              using one of supported methods:
              http://django-reversion.readthedocs.org/en/latest/api.html#registering-models-with-django-reversion
    """

    def get_version_fields(self):
        """ Get field that are tracked in object history versions. """
        options = reversion._get_options(self)
        return options.fields or [f.name for f in self._meta.fields if f not in options.exclude]

    def _is_version_duplicate(self):
        """ Define should new version be created for object or no.

            Reasons to provide custom check instead of default `ignore_revision_duplicates`:
             - no need to compare all revisions - it is OK if right object version exists in any revision;
             - need to compare object attributes (not serialized data) to avoid
               version creation on wrong <float> vs <int> comparison;
        """
        if self.id is None:
            return False
        try:
            latest_version = Version.objects.get_for_object(self).latest('revision__date_created')
        except Version.DoesNotExist:
            return False
        latest_version_object = latest_version._object_version.object
        fields = self.get_version_fields()
        return all([getattr(self, f) == getattr(latest_version_object, f) for f in fields])

    def save(self, **kwargs):
        if self._is_version_duplicate():
            return super(ReversionMixin, self).save(**kwargs)
        with reversion.create_revision():
            return super(ReversionMixin, self).save(**kwargs)


# XXX: consider renaming it to AffinityMixin
class DescendantMixin:
    """ Mixin to provide child-parent relationships.
        Each related model can provide list of its parents/children.
    """

    def get_parents(self):
        """ Return list instance parents. """
        return []

    def get_children(self):
        """ Return list instance children. """
        return []

    def get_ancestors(self):
        """ Get all unique instance ancestors """
        ancestors = list(self.get_parents())
        ancestor_unique_attributes = set([(a.__class__, a.id) for a in ancestors])
        ancestors_with_parents = [a for a in ancestors if isinstance(a, DescendantMixin)]
        for ancestor in ancestors_with_parents:
            for parent in ancestor.get_ancestors():
                if (parent.__class__, parent.id) not in ancestor_unique_attributes:
                    ancestors.append(parent)
        return ancestors

    def get_descendants(self):
        def traverse(obj):
            for child in obj.get_children():
                yield child
                for baby in child.get_descendants():
                    yield baby
        return list(set(traverse(self)))


class AbstractFieldTracker(FieldTracker):
    """
    Workaround for abstract models
    https://gist.github.com/sbnoemi/7618916
    """

    def finalize_class(self, sender, name, **kwargs):
        self.name = name
        self.attname = '_%s' % name
        if not hasattr(sender, name):
            super(AbstractFieldTracker, self).finalize_class(sender, **kwargs)


class BackendModelMixin:
    """
    Represents model that is connected to backend object.

    This model cannot be created or updated via admin, because we do not support queries to backend from admin interface.
    """

    @classmethod
    def get_backend_fields(cls):
        """
        Returns a list of fields that are handled on backend.
        """
        return ()



import django_filters

from waldur_core.core import filters as core_filters

from . import models


class AnsibleJobsFilter(django_filters.FilterSet):
    name = django_filters.CharFilter(lookup_expr='icontains')
    project = core_filters.URLFilter(view_name='project-detail', field_name='service_project_link__project__uuid')
    project_uuid = django_filters.UUIDFilter(field_name='service_project_link__project__uuid')

    class Meta:
        model = models.Job
        fields = ['state']
        o = django_filters.OrderingFilter(fields=(
            'name',
            'state',
            'created',
            'modified',
        ))



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import ugettext_lazy as _

from waldur_core.core.fields import JSONField
from waldur_core.core import models as core_models
from waldur_core.structure import models as structure_models

from . import managers


class ZabbixService(structure_models.Service):
    projects = models.ManyToManyField(
        structure_models.Project, related_name='zabbix_services', through='ZabbixServiceProjectLink')

    @classmethod
    def get_url_name(cls):
        return 'zabbix'


class ZabbixServiceProjectLink(structure_models.ServiceProjectLink):
    service = models.ForeignKey(on_delete=models.CASCADE, to=ZabbixService)

    @classmethod
    def get_url_name(cls):
        return 'zabbix-spl'


class Host(structure_models.NewResource):
    VISIBLE_NAME_MAX_LENGTH = 64

    # list of items that are added as monitoring items to hosts scope.
    # parameters:
    #  zabbix_item_key - Zabbix item key,
    #  monitoring_item_name - name of monitoring item that will be attached to host scope,
    #  after_creation_update - True if monitoring item need to be updated frequently after host creation,
    #  after_creation_update_terminate_values - stop after_creation_update if monitoring item value is one of
    #                                           terminated values.
    MONITORING_ITEMS_CONFIGS = [
        {
            'zabbix_item_key': 'application.status',
            'monitoring_item_name': 'application_state',
            'after_creation_update': True,
            'after_creation_update_terminate_values': ['1'],
        }
    ]

    class Statuses:
        MONITORED = '0'
        UNMONITORED = '1'

        CHOICES = ((MONITORED, 'monitored'), (UNMONITORED, 'unmonitored'))

    service_project_link = models.ForeignKey(ZabbixServiceProjectLink, related_name='hosts', on_delete=models.PROTECT)
    visible_name = models.CharField(_('visible name'), max_length=VISIBLE_NAME_MAX_LENGTH)
    interface_parameters = JSONField(blank=True)
    host_group_name = models.CharField(_('host group name'), max_length=64, blank=True)
    error = models.CharField(max_length=500, blank=True, help_text='Error text if Zabbix agent is unavailable.')
    status = models.CharField(max_length=30, choices=Statuses.CHOICES, default=Statuses.MONITORED)
    templates = models.ManyToManyField('Template', related_name='hosts')

    content_type = models.ForeignKey(on_delete=models.CASCADE, to=ContentType, null=True)
    object_id = models.PositiveIntegerField(null=True)
    scope = GenericForeignKey('content_type', 'object_id')

    objects = managers.HostManager('scope')

    def __str__(self):
        return '%s (%s)' % (self.name, self.visible_name)

    @classmethod
    def get_url_name(cls):
        return 'zabbix-host'

    def clean(self):
        # It is impossible to mark service and name unique together at DB level, because host is connected with service
        # through SPL.
        same_service_hosts = Host.objects.filter(service_project_link__service=self.service_project_link.service)
        if same_service_hosts.filter(name=self.name).exclude(pk=self.pk).exists():
            raise ValidationError(
                'Host with name "%s" already exists at this service. Host name should be unique.' % self.name)
        if same_service_hosts.filter(visible_name=self.visible_name).exclude(pk=self.pk).exists():
            raise ValidationError('Host with visible_name "%s" already exists at this service.'
                                  ' Host name should be unique.' % self.visible_name)

    @classmethod
    def get_visible_name_from_scope(cls, scope):
        """ Generate visible name based on host scope """
        return ('%s-%s' % (scope.uuid.hex, scope.name))[:64]


# Zabbix host name max length - 64
Host._meta.get_field('name').max_length = 64


class Template(structure_models.ServiceProperty):
    parents = models.ManyToManyField('Template', related_name='children')

    @classmethod
    def get_url_name(cls):
        return 'zabbix-template'


class Item(models.Model):
    class ValueTypes:
        FLOAT = 0
        CHAR = 1
        LOG = 2
        INTEGER = 3
        TEXT = 4

        CHOICES = (
            (FLOAT, 'Numeric (float)'),
            (CHAR, 'Character'),
            (LOG, 'Log'),
            (INTEGER, 'Numeric (unsigned)'),
            (TEXT, 'Text')
        )

    key = models.CharField(max_length=255)
    name = models.CharField(max_length=255)
    template = models.ForeignKey(on_delete=models.CASCADE, to=Template, related_name='items')
    backend_id = models.CharField(max_length=64)
    value_type = models.IntegerField(choices=ValueTypes.CHOICES)
    units = models.CharField(max_length=255)
    history = models.IntegerField()
    delay = models.IntegerField()

    def is_byte(self):
        return self.units == 'B'

    def __str__(self):
        return self.name


class Trigger(structure_models.ServiceProperty):
    template = models.ForeignKey(on_delete=models.CASCADE, to=Template, related_name='triggers')
    # https://www.zabbix.com/documentation/3.4/manual/api/reference/trigger/object
    priority = models.IntegerField(default=0)

    class Priority:
        DEFAULT = 0
        INFORMATION = 1
        WARNING = 2
        AVERAGE = 3
        HIGH = 4
        DISASTER = 5

        CHOICES = (
            (DEFAULT, 'Default'),
            (INFORMATION, 'Information'),
            (WARNING, 'Warning'),
            (AVERAGE, 'Average'),
            (HIGH, 'High'),
            (DISASTER, 'Disaster'),
        )

    class AcknowledgeStatus:
        SOME_EVENTS_UNACKNOWLEDGED = 1
        LAST_EVENT_UNACKNOWLEDGED = 2
        ALL_EVENTS_ACKNOWLEDGED = 3

        CHOICES = (
            (SOME_EVENTS_UNACKNOWLEDGED, 'Some events unacknowledged'),
            (LAST_EVENT_UNACKNOWLEDGED, 'Last event unacknowledged'),
            (ALL_EVENTS_ACKNOWLEDGED, 'All events unacknowledged'),
        )

    class Value:
        OK = 0
        PROBLEM = 1

        CHOICES = (
            (OK, 'OK'),
            (PROBLEM, 'Problem'),
        )

    @classmethod
    def get_url_name(cls):
        return 'zabbix-trigger'

    def __str__(self):
        return '%s-%s | %s' % (self.template.name, self.name, self.settings)


# Zabbix trigger name max length - 255
Trigger._meta.get_field('name').max_length = 255


class ITService(structure_models.NewResource):
    class Algorithm:
        SKIP = 0
        ANY = 1
        ALL = 2

        CHOICES = (
            (SKIP, 'do not calculate'),
            (ANY, 'problem, if at least one child has a problem'),
            (ALL, 'problem, if all children have problems')
        )

    service_project_link = models.ForeignKey(
        ZabbixServiceProjectLink, related_name='itservices', on_delete=models.PROTECT)
    host = models.ForeignKey(on_delete=models.CASCADE, to=Host, related_name='itservices', blank=True, null=True)
    is_main = models.BooleanField(
        default=True, help_text='Main IT service SLA will be added to hosts resource as monitoring item.')

    algorithm = models.PositiveSmallIntegerField(choices=Algorithm.CHOICES, default=Algorithm.SKIP)
    sort_order = models.PositiveSmallIntegerField(default=1)
    agreed_sla = models.DecimalField(max_digits=6, decimal_places=4, null=True, blank=True)

    backend_trigger_id = models.CharField(max_length=64, null=True, blank=True)
    trigger = models.ForeignKey(on_delete=models.CASCADE, to=Trigger, null=True, blank=True)

    class Meta:
        unique_together = ('host', 'is_main')

    @classmethod
    def get_url_name(cls):
        return 'zabbix-itservice'


class SlaHistory(models.Model):
    itservice = models.ForeignKey(on_delete=models.CASCADE, to=ITService)
    period = models.CharField(max_length=10)
    value = models.DecimalField(max_digits=11, decimal_places=4, null=True, blank=True)

    class Meta:
        verbose_name = 'SLA history'
        verbose_name_plural = 'SLA histories'
        unique_together = ('itservice', 'period')

    def __str__(self):
        return 'SLA for %s during %s: %s' % (self.itservice, self.period, self.value)


class SlaHistoryEvent(models.Model):
    EVENTS = (
        ('U', 'DOWN'),
        ('D', 'UP'),
    )

    history = models.ForeignKey(on_delete=models.CASCADE, to=SlaHistory, related_name='events')
    timestamp = models.IntegerField()
    state = models.CharField(max_length=1, choices=EVENTS)

    def __str__(self):
        return '%s - %s' % (self.timestamp, self.state)


class UserGroup(structure_models.ServiceProperty):
    @classmethod
    def get_url_name(cls):
        return 'zabbix-user-group'

    def get_backend(self):
        return self.settings.get_backend()


class User(core_models.StateMixin, structure_models.ServiceProperty):
    class Types:
        DEFAULT = '1'
        ADMIN = '2'
        SUPERADMIN = '3'

        CHOICES = ((DEFAULT, 'default'), (ADMIN, 'admin'), (SUPERADMIN, 'superadmin'))

    alias = models.CharField(max_length=150)
    surname = models.CharField(max_length=150)
    type = models.CharField(max_length=30, choices=Types.CHOICES, default=Types.DEFAULT)
    groups = models.ManyToManyField(UserGroup, related_name='users')
    # password can be blank if user was pulled from zabbix, not created through NC
    password = models.CharField(max_length=150, blank=True)
    # phone is NC-only field
    phone = models.CharField(max_length=30, blank=True)

    class Meta:
        unique_together = ('alias', 'settings')

    def __str__(self):
        return '%s | %s' % (self.alias, self.settings)

    @classmethod
    def get_url_name(cls):
        return 'zabbix-user'

    def get_backend(self):
        return self.settings.get_backend()



from waldur_core.core import WaldurExtension


class JiraExtension(WaldurExtension):

    class Settings:
        WALDUR_JIRA = {
            'COMMENT_TEMPLATE': '{body}\n\n_(added by {user.full_name} [{user.username}] via G-Cloud Portal)_',
            'ISSUE_TEMPLATE': {
                'RESOURCE_INFO': '\nAffected resource: {resource}\n'
            },
            'ISSUE': {
                'resolution_sla_field': 'Time to resolution',
            },
            'ISSUE_IMPORT_LIMIT': 10
        }

    @staticmethod
    def django_app():
        return 'waldur_jira'

    @staticmethod
    def django_urls():
        from .urls import urlpatterns
        return urlpatterns

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    # @staticmethod
    # def celery_tasks():
    #     return {
    #         'waldur-import-jira-projects': {
    #             'task': 'waldur_jira.ImportProjects',
    #             'schedule': timedelta(minutes=1),
    #             'args': (),
    #         },
    #     }



from django.apps import AppConfig
from django.db.models import signals


class RancherConfig(AppConfig):
    name = 'waldur_rancher'
    verbose_name = 'Rancher'
    service_name = 'Rancher'

    def ready(self):
        from waldur_core.structure import SupportedServices

        from waldur_openstack.openstack_tenant.models import Instance

        from .backend import RancherBackend
        from . import handlers, models, signals as rancher_signals

        SupportedServices.register_backend(RancherBackend)

        rancher_signals.rancher_user_created.connect(
            handlers.notify_create_user,
            sender=models.RancherUser,
            dispatch_uid='waldur_rancher.notify_create_user',
        )

        signals.post_delete.connect(
            handlers.delete_catalog_when_cluster_is_deleted,
            sender=models.Cluster,
            dispatch_uid='waldur_rancher.delete_catalog_when_cluster_is_deleted',
        )

        signals.post_delete.connect(
            handlers.delete_node_if_related_instance_has_been_deleted,
            sender=Instance,
            dispatch_uid='waldur_rancher.delete_node_if_related_instance_has_been_deleted',
        )

        signals.post_delete.connect(
            handlers.delete_cluster_if_all_related_nodes_have_been_deleted,
            sender=models.Node,
            dispatch_uid='waldur_rancher.delete_cluster_if_all_related_nodes_have_been_deleted',
        )

        signals.post_save.connect(
            handlers.set_error_state_for_node_if_related_instance_deleting_is_failed,
            sender=Instance,
            dispatch_uid='waldur_rancher.set_error_state_for_node_if_related_instance_deleting_is_failed',
        )

        signals.post_save.connect(
            handlers.set_error_state_for_cluster_if_related_node_deleting_is_failed,
            sender=models.Node,
            dispatch_uid='waldur_rancher.set_error_state_for_cluster_if_related_node_deleting_is_failed',
        )



from django.db import models
from django.utils.translation import ugettext_lazy as _

from waldur_core.quotas.fields import CounterQuotaField
from waldur_core.quotas.models import QuotaModelMixin
from waldur_core.structure import models as structure_models


class DigitalOceanService(structure_models.Service):
    projects = models.ManyToManyField(
        structure_models.Project, related_name='digitalocean_services', through='DigitalOceanServiceProjectLink')

    class Meta(structure_models.Service.Meta):
        verbose_name = _('DigitalOcean provider')
        verbose_name_plural = _('DigitalOcean providers')

    class Quotas(QuotaModelMixin.Quotas):
        droplet_count = CounterQuotaField(
            default_limit=50,
            target_models=lambda: [Droplet],
            path_to_scope='service_project_link.service'
        )

    @classmethod
    def get_url_name(cls):
        return 'digitalocean'


class DigitalOceanServiceProjectLink(structure_models.CloudServiceProjectLink):
    service = models.ForeignKey(on_delete=models.CASCADE, to=DigitalOceanService)

    class Meta(structure_models.CloudServiceProjectLink.Meta):
        verbose_name = _('DigitalOcean provider project link')
        verbose_name_plural = _('DigitalOcean provider project links')

    @classmethod
    def get_url_name(cls):
        return 'digitalocean-spl'


class Region(structure_models.GeneralServiceProperty):
    @classmethod
    def get_url_name(cls):
        return 'digitalocean-region'


class Image(structure_models.GeneralServiceProperty):
    regions = models.ManyToManyField(Region)
    distribution = models.CharField(max_length=100)
    type = models.CharField(max_length=100)
    is_official = models.BooleanField(default=False, help_text=_('Is image provided by DigitalOcean'))
    min_disk_size = models.PositiveIntegerField(
        null=True, help_text=_('Minimum disk required for a size to use this image'))
    created_at = models.DateTimeField(null=True)

    @property
    def is_ssh_key_mandatory(self):
        MANDATORY = 'Ubuntu', 'FreeBSD', 'CoreOS'
        return self.distribution in MANDATORY

    def __str__(self):
        return '{} {} ({})'.format(self.name, self.distribution, self.type)

    @classmethod
    def get_url_name(cls):
        return 'digitalocean-image'

    @classmethod
    def get_backend_fields(cls):
        return super(Image, cls).get_backend_fields() + ('type', 'distribution', 'is_official', 'min_disk_size',
                                                         'created_at')


class Size(structure_models.GeneralServiceProperty):
    regions = models.ManyToManyField(Region)

    cores = models.PositiveSmallIntegerField(help_text=_('Number of cores in a VM'))
    ram = models.PositiveIntegerField(help_text=_('Memory size in MiB'))
    disk = models.PositiveIntegerField(help_text=_('Disk size in MiB'))
    transfer = models.PositiveIntegerField(help_text=_('Amount of transfer bandwidth in MiB'))
    price = models.DecimalField(_('Hourly price rate'), default=0, max_digits=11, decimal_places=5)

    @classmethod
    def get_url_name(cls):
        return 'digitalocean-size'

    @classmethod
    def get_backend_fields(cls):
        return super(Size, cls).get_backend_fields() + ('cores', 'ram', 'disk', 'transfer', 'price')


class Droplet(structure_models.VirtualMachine):
    service_project_link = models.ForeignKey(
        DigitalOceanServiceProjectLink, related_name='droplets', on_delete=models.PROTECT)
    transfer = models.PositiveIntegerField(default=0, help_text=_('Amount of transfer bandwidth in MiB'))
    ip_address = models.GenericIPAddressField(null=True, protocol='IPv4', blank=True)
    region_name = models.CharField(max_length=150, blank=True)
    size_name = models.CharField(max_length=150, blank=True)

    def increase_backend_quotas_usage(self, validate=True):
        spl = self.service_project_link
        spl.add_quota_usage(spl.Quotas.storage, self.disk, validate=validate)
        spl.add_quota_usage(spl.Quotas.ram, self.ram, validate=validate)
        spl.add_quota_usage(spl.Quotas.vcpu, self.cores, validate=validate)

    def decrease_backend_quotas_usage(self):
        self.service_project_link.add_quota_usage(self.service_project_link.Quotas.storage, -self.disk)
        self.service_project_link.add_quota_usage(self.service_project_link.Quotas.ram, -self.ram)
        self.service_project_link.add_quota_usage(self.service_project_link.Quotas.vcpu, -self.cores)

    @property
    def external_ips(self):
        return [self.ip_address]

    @property
    def internal_ips(self):
        return []

    @classmethod
    def get_url_name(cls):
        return 'digitalocean-droplet'

    @classmethod
    def get_backend_fields(cls):
        return super(Droplet, cls).get_backend_fields() + ('state', 'runtime_state', 'image_name')



from django.contrib import admin

from waldur_core.users import models


class InvitationAdmin(admin.ModelAdmin):
    list_display = ('email', 'state', 'created', 'created_by', 'customer', 'project')
    list_filter = ('state', 'created')
    search_fields = ('email', 'customer__name')


admin.site.register(models.Invitation, InvitationAdmin)



from waldur_core.core.models import User, SshPublicKey
from waldur_core.logging.loggers import EventLogger, event_logger


class AuthEventLogger(EventLogger):
    user = User
    username = str

    class Meta:
        event_types = ('auth_logged_in_with_username',
                       'auth_login_failed_with_username',
                       'auth_logged_out')
        event_groups = {'users': event_types}
        nullable_fields = ['user', 'username']

    @staticmethod
    def get_scopes(event_context):
        if 'user' in event_context:
            return {event_context['user']}
        else:
            return set()


class UserEventLogger(EventLogger):
    affected_user = User

    class Meta:
        permitted_objects_uuids = staticmethod(lambda user: {'user_uuid': [user.uuid.hex]})
        event_types = ('user_creation_succeeded',
                       'user_update_succeeded',
                       'user_deletion_succeeded',
                       'user_password_updated',
                       'user_token_lifetime_updated',
                       'user_activated',
                       'user_deactivated',
                       'user_profile_changed')
        event_groups = {
            'users': event_types,
        }

    @staticmethod
    def get_scopes(event_context):
        return {event_context['affected_user']}


class TokenEventLogger(EventLogger):
    affected_user = User

    class Meta:
        event_types = ('token_created',)

    @staticmethod
    def get_scopes(event_context):
        return {event_context['affected_user']}


class SshPublicKeyEventLogger(EventLogger):
    ssh_key = SshPublicKey
    user = User

    class Meta:
        event_types = ('ssh_key_creation_succeeded',
                       'ssh_key_deletion_succeeded')
        event_groups = {
            'ssh': event_types,
            'users': event_types,
        }

    @staticmethod
    def get_scopes(event_context):
        return {event_context['user']}


event_logger.register('auth', AuthEventLogger)
event_logger.register('user', UserEventLogger)
event_logger.register('sshkey', SshPublicKeyEventLogger)
event_logger.register('token', TokenEventLogger)



from django.apps import AppConfig
from django.db.models import signals


class MarketplaceSlurmConfig(AppConfig):
    name = 'waldur_mastermind.marketplace_slurm'
    verbose_name = 'Marketplace SLURM'

    def ready(self):
        from waldur_mastermind.marketplace.plugins import Component, manager
        from waldur_mastermind.marketplace import models as marketplace_models
        from waldur_mastermind.marketplace import handlers as marketplace_handlers
        from waldur_mastermind.marketplace_slurm import PLUGIN_NAME
        from waldur_slurm import models as slurm_models
        from waldur_slurm.apps import SlurmConfig

        from . import handlers, processor

        signals.post_save.connect(
            handlers.synchronize_slurm_package,
            sender=marketplace_models.PlanComponent,
            dispatch_uid='waldur_mastermind.marketpace_slurm.synchronize_slurm_package',
        )

        signals.post_save.connect(
            handlers.create_slurm_usage,
            sender=slurm_models.AllocationUsage,
            dispatch_uid='waldur_mastermind.marketpace_slurm.create_slurm_usage',
        )

        signals.post_save.connect(
            handlers.update_component_quota,
            sender=slurm_models.Allocation,
            dispatch_uid='waldur_mastermind.marketpace_slurm.update_component_quota',
        )

        signals.post_save.connect(
            handlers.add_component_usage,
            sender=marketplace_models.ComponentUsage,
            dispatch_uid='marketplace_slurm.handlers.add_component_usage',
        )

        marketplace_handlers.connect_resource_handlers(slurm_models.Allocation)
        marketplace_handlers.connect_resource_metadata_handlers(slurm_models.Allocation)

        USAGE = marketplace_models.OfferingComponent.BillingTypes.USAGE
        manager.register(PLUGIN_NAME,
                         create_resource_processor=processor.CreateAllocationProcessor,
                         delete_resource_processor=processor.DeleteAllocationProcessor,
                         components=(
                             Component(type='cpu', name='CPU', measured_unit='hours', billing_type=USAGE),
                             Component(type='gpu', name='GPU', measured_unit='hours', billing_type=USAGE),
                             Component(type='ram', name='RAM', measured_unit='GB', billing_type=USAGE),
                         ),
                         service_type=SlurmConfig.service_name)



from waldur_core.core import WaldurExtension


class MarketplaceScriptExtension(WaldurExtension):

    class Settings:
        WALDUR_MARKETPLACE_SCRIPT = {
            # See also: https://docker-py.readthedocs.io/en/stable/client.html#docker.client.DockerClient
            'DOCKER_CLIENT': {
                'base_url': None,
            },
            # See also: https://docker-py.readthedocs.io/en/stable/containers.html#docker.models.containers.ContainerCollection.run
            'DOCKER_RUN_OPTIONS': {
                'mem_limit': '64m',
            },
            # Key is command to execute script, value is image name.
            'DOCKER_IMAGES': {
                'python': 'python:3.7-alpine',
                'sh': 'alpine:3.10.0',
            },
        }

    @staticmethod
    def django_app():
        return 'waldur_mastermind.marketplace_script'

    @staticmethod
    def is_assembly():
        return True

    @staticmethod
    def celery_tasks():
        from datetime import timedelta
        return {
            'waldur-marketplace-script-pull-resources': {
                'task': 'waldur_marketplace_script.pull_resources',
                'schedule': timedelta(hours=1),
                'args': (),
            },
        }



from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from django.utils.dateparse import datetime_re
from rest_framework.serializers import ValidationError

from waldur_mastermind.marketplace import processors
from waldur_mastermind.marketplace import models as marketplace_models

from .utils import TimePeriod, is_interval_in_schedules


class BookingCreateProcessor(processors.BaseOrderItemProcessor):
    def process_order_item(self, user):
        with transaction.atomic():
            resource = marketplace_models.Resource(
                project=self.order_item.order.project,
                offering=self.order_item.offering,
                plan=self.order_item.plan,
                limits=self.order_item.limits,
                attributes=self.order_item.attributes,
                name=self.order_item.attributes.get('name') or '',
                state=marketplace_models.Resource.States.CREATING,
            )
            resource.init_cost()
            resource.save()
            resource.init_quotas()
            self.order_item.resource = resource
            self.order_item.save(update_fields=['resource'])

    def validate_order_item(self, request):
        schedules = self.order_item.attributes.get('schedules')

        # We check that the schedule is set.
        if not schedules:
            raise ValidationError(_('Schedules are required.'))

        if not len(schedules):
            raise ValidationError(_('Schedules are required.'))

        for period in schedules:
            try:
                start = period['start']
                end = period['end']
            except KeyError:
                raise ValidationError(_('Key \'start\' or \'end\' does not exist in schedules item.'))

            for value in [start, end]:
                match = datetime_re.match(value)
                kw = match.groupdict()
                if list(filter(lambda x: not kw[x], ['hour', 'month', 'second', 'year', 'tzinfo', 'day', 'minute'])):
                    raise ValidationError(_('The value %s does not match the format.') % value)

        # Check that the schedule is available for the offering.
        offering = self.order_item.offering
        offering_schedules = offering.attributes.get('schedules', [])

        for period in schedules:
            if not is_interval_in_schedules(TimePeriod(period['start'], period['end']),
                                            [TimePeriod(i['start'], i['end']) for i in offering_schedules]):
                raise ValidationError(_('Time period from %s to %s is not available for selected offering.') %
                                      (period['start'], period['end']))

        # Check that there are no other bookings.
        bookings = []

        for resource in marketplace_models.Resource.objects.filter(offering=offering,
                                                                   state=marketplace_models.Resource.States.OK):
            for period in resource.attributes.get('schedules', []):
                bookings.append(TimePeriod(period['start'], period['end']))

        for period in schedules:
            if is_interval_in_schedules(TimePeriod(period['start'], period['end']), bookings):
                raise ValidationError(_('Time period from %s to %s is not available.') %
                                      (period['start'], period['end']))


class BookingDeleteProcessor(processors.DeleteResourceProcessor):
    pass



from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import response, viewsets, permissions, status, decorators, mixins

from waldur_core.core import filters as core_filters, permissions as core_permissions
from waldur_core.core.managers import SummaryQuerySet
from waldur_core.logging import models, serializers, filters, utils
from waldur_core.logging.loggers import get_event_groups


class EventViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = models.Event.objects.all()
    permission_classes = (permissions.IsAuthenticated, core_permissions.IsAdminOrReadOnly)
    serializer_class = serializers.EventSerializer
    filter_backends = (DjangoFilterBackend, filters.EventFilterBackend)
    filterset_class = filters.EventFilter

    @decorators.action(detail=False)
    def count(self, request, *args, **kwargs):
        """
        To get a count of events - run **GET** against */api/events/count/* as authenticated user.
        Endpoint support same filters as events list.

        Response example:

        .. code-block:: javascript

            {"count": 12321}
        """

        self.queryset = self.filter_queryset(self.get_queryset())
        return response.Response({'count': self.queryset.count()}, status=status.HTTP_200_OK)

    @decorators.action(detail=False)
    def scope_types(self, request, *args, **kwargs):
        """ Returns a list of scope types acceptable by events filter. """
        return response.Response(utils.get_scope_types_mapping().keys())

    @decorators.action(detail=False)
    def event_groups(self, request, *args, **kwargs):
        """
        Returns a list of groups with event types.
        Group is used in exclude_features query param.
        """
        return response.Response(get_event_groups())


class BaseHookViewSet(viewsets.ModelViewSet):
    """
    Hooks API allows user to receive event notifications via different channel, like email or webhook.
    To get a list of all your hooks, run **GET** against */api/hooks/* as an authenticated user.
    """
    filter_backends = (core_filters.StaffOrUserFilter, DjangoFilterBackend)
    lookup_field = 'uuid'


class WebHookViewSet(BaseHookViewSet):
    queryset = models.WebHook.objects.all()
    filterset_class = filters.WebHookFilter
    serializer_class = serializers.WebHookSerializer

    def create(self, request, *args, **kwargs):
        """
        To create new web hook issue **POST** against */api/hooks-web/* as an authenticated user.
        You should specify list of event_types or event_groups.

        Example of a request:

        .. code-block:: http

            POST /api/hooks-web/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "event_types": ["resource_start_succeeded"],
                "event_groups": ["users"],
                "destination_url": "http://example.com/"
            }

        When hook is activated, **POST** request is issued against destination URL with the following data:

        .. code-block:: javascript

            {
                "timestamp": "2015-07-14T12:12:56.000000",
                "message": "Customer ABC LLC has been updated.",
                "type": "customer_update_succeeded",
                "context": {
                    "user_native_name": "Walter Lebrowski",
                    "customer_contact_details": "",
                    "user_username": "Walter",
                    "user_uuid": "1c3323fc4ae44120b57ec40dea1be6e6",
                    "customer_uuid": "4633bbbb0b3a4b91bffc0e18f853de85",
                    "ip_address": "8.8.8.8",
                    "user_full_name": "Walter Lebrowski",
                    "customer_abbreviation": "ABC LLC",
                    "customer_name": "ABC LLC"
                },
                "levelname": "INFO"
            }

        Note that context depends on event type.
        """
        return super(WebHookViewSet, self).create(request, *args, **kwargs)


class EmailHookViewSet(BaseHookViewSet):
    queryset = models.EmailHook.objects.all()
    filterset_class = filters.EmailHookFilter
    serializer_class = serializers.EmailHookSerializer

    def create(self, request, *args, **kwargs):
        """
        To create new email hook issue **POST** against */api/hooks-email/* as an authenticated user.
        You should specify list of event_types or event_groups.

        Example of a request:

        .. code-block:: http

            POST /api/hooks-email/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "event_types": ["openstack_instance_start_succeeded"],
                "event_groups": ["users"],
                "email": "test@example.com"
            }

        You may temporarily disable hook without deleting it by issuing following **PATCH** request against hook URL:

        .. code-block:: javascript

            {
                "is_active": "false"
            }
        """
        return super(EmailHookViewSet, self).create(request, *args, **kwargs)


class PushHookViewSet(BaseHookViewSet):
    queryset = models.PushHook.objects.all()
    filterset_class = filters.PushHookFilter
    serializer_class = serializers.PushHookSerializer

    def create(self, request, *args, **kwargs):
        """
        To create new push hook issue **POST** against */api/hooks-push/* as an authenticated user.
        You should specify list of event_types or event_groups.

        Example of a request:

        .. code-block:: http

            POST /api/hooks-push/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "event_types": ["resource_start_succeeded"],
                "event_groups": ["users"],
                "type": "Android"
            }

        You may temporarily disable hook without deleting it by issuing following **PATCH** request against hook URL:

        .. code-block:: javascript

            {
                "is_active": "false"
            }
        """
        return super(PushHookViewSet, self).create(request, *args, **kwargs)


class HookSummary(mixins.ListModelMixin, viewsets.GenericViewSet):
    """
    Use */api/hooks/* to get a list of all the hooks of any type that a user can see.
    """
    serializer_class = serializers.SummaryHookSerializer
    filter_backends = (core_filters.StaffOrUserFilter, filters.HookSummaryFilterBackend)

    def get_queryset(self):
        return SummaryQuerySet(models.BaseHook.get_all_models())



import collections
import logging
import re

from django.core.exceptions import ObjectDoesNotExist
from django.conf import settings as django_settings
from django.db import transaction
from django.db.models import Q
from django.utils import timezone
from django.utils.translation import ugettext, ugettext_lazy as _
import pytz
from rest_framework import serializers
from rest_framework.reverse import reverse

from waldur_core.core import (serializers as core_serializers,
                              utils as core_utils,
                              signals as core_signals)
from waldur_core.quotas import serializers as quotas_serializers
from waldur_core.structure import serializers as structure_serializers, SupportedServices
from waldur_core.structure import models as structure_models
from waldur_core.structure.permissions import _has_admin_access
from waldur_openstack.openstack import models as openstack_models
from waldur_openstack.openstack import serializers as openstack_serializers
from waldur_openstack.openstack_base.backend import OpenStackBackendError
from waldur_openstack.openstack_base.serializers import BaseVolumeTypeSerializer
from waldur_openstack.openstack_tenant.utils import get_valid_availability_zones

from . import models, fields

logger = logging.getLogger(__name__)


class ServiceSerializer(core_serializers.ExtraFieldOptionsMixin,
                        core_serializers.RequiredFieldsMixin,
                        structure_serializers.BaseServiceSerializer):
    SERVICE_ACCOUNT_FIELDS = {
        'backend_url': _('Keystone auth URL (e.g. http://keystone.example.com:5000/v3)'),
        'domain': _('Tenant domain'),
        'username': _('Tenant user username'),
        'password': _('Tenant user password'),
    }
    SERVICE_ACCOUNT_EXTRA_FIELDS = {
        'tenant_id': _('Tenant ID in OpenStack'),
        'availability_zone': _('Default availability zone for provisioned instances'),
        'flavor_exclude_regex': _('Flavors matching this regex expression will not be pulled from the backend.'),
        'external_network_id': _('It is used to automatically assign floating IP to your virtual machine.'),
        'console_type': _('The type of remote console. '
                          'The valid values are novnc, xvpvnc, rdp-html5, '
                          'spice-html5, serial, and webmks.'),
        'config_drive': _('Indicates whether a config drive enables metadata injection'),
    }

    # Expose service settings quotas as service quotas as a temporary workaround.
    # It is needed in order to render quotas table in service provider details dialog.
    quotas = quotas_serializers.BasicQuotaSerializer(many=True, read_only=True, source='settings.quotas')

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.OpenStackTenantService
        required_fields = ('backend_url', 'username', 'password', 'tenant_id',)
        extra_field_options = {
            'backend_url': {
                'label': 'API URL',
                'default_value': 'http://keystone.example.com:5000/v3',
            },
            'tenant_id': {
                'label': 'Tenant ID',
            },
            'availability_zone': {
                'placeholder': 'default',
            },
            'external_network_id': {
                'required': True,
            },
            'console_type': {
                'default_value': 'novnc',
            },
        }


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):
    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.OpenStackTenantServiceProjectLink
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'openstacktenant-detail'},
        }


class BaseAvailabilityZoneSerializer(structure_serializers.BasePropertySerializer):
    settings = serializers.HyperlinkedRelatedField(
        queryset=structure_models.ServiceSettings.objects.all(),
        view_name='servicesettings-detail',
        lookup_field='uuid',
        allow_null=True,
        required=False,
    )

    class Meta(structure_serializers.BasePropertySerializer.Meta):
        fields = ('url', 'uuid', 'name', 'settings', 'available')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'settings': {'lookup_field': 'uuid'},
        }


class ImageSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.Image
        fields = ('url', 'uuid', 'name', 'settings', 'min_disk', 'min_ram',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'settings': {'lookup_field': 'uuid'},
        }


class FlavorSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.Flavor
        fields = ('url', 'uuid', 'name', 'settings', 'cores', 'ram', 'disk',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'settings': {'lookup_field': 'uuid'},
        }


class UsageStatsSerializer(serializers.Serializer):
    shared = serializers.BooleanField()
    service_provider = serializers.ListField(child=serializers.CharField())


class NetworkSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.Network
        fields = ('url', 'uuid', 'name',
                  'type', 'is_external', 'segmentation_id', 'subnets')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'settings': {'lookup_field': 'uuid'},
            'subnets': {'lookup_field': 'uuid', 'view_name': 'openstacktenant-subnet-detail'}
        }


class SubNetSerializer(structure_serializers.BasePropertySerializer):
    dns_nameservers = serializers.JSONField(read_only=True)
    allocation_pools = serializers.JSONField(read_only=True)
    network_name = serializers.ReadOnlyField(source='network.name')

    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.SubNet
        fields = ('url', 'uuid', 'name',
                  'cidr', 'gateway_ip', 'allocation_pools', 'ip_version', 'enable_dhcp',
                  'dns_nameservers', 'network', 'network_name')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'settings': {'lookup_field': 'uuid'},
            'network': {'lookup_field': 'uuid', 'view_name': 'openstacktenant-network-detail'},
        }


class FloatingIPSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.FloatingIP
        fields = ('url', 'uuid', 'settings', 'address', 'runtime_state', 'is_booked',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'settings': {'lookup_field': 'uuid'},
        }


class SecurityGroupSerializer(structure_serializers.BasePropertySerializer):
    rules = serializers.SerializerMethodField()

    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.SecurityGroup
        fields = ('url', 'uuid', 'name', 'settings', 'description', 'rules')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'settings': {'lookup_field': 'uuid'},
        }

    def get_rules(self, security_group):
        rules = []
        for rule in security_group.rules.all():
            rules.append({
                'protocol': rule.protocol,
                'from_port': rule.from_port,
                'to_port': rule.to_port,
                'cidr': rule.cidr,
            })
        return rules


class VolumeImportableSerializer(core_serializers.AugmentedSerializerMixin,
                                 serializers.HyperlinkedModelSerializer):
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-spl-detail',
        queryset=models.OpenStackTenantServiceProjectLink.objects.all(),
        write_only=True)

    type = serializers.SerializerMethodField()
    extra = serializers.SerializerMethodField()

    def get_type(self, obj):
        return SupportedServices.get_name_for_model(models.Volume)

    def get_extra(self, volume):
        return [
            {
                'name': 'Is bootable',
                'value': volume.bootable,
            },
            {
                'name': 'Size',
                'value': volume.size,
            },
            {
                'name': 'Description',
                'value': volume.description,
            },
            {
                'name': 'Device',
                'value': volume.device,
            },
            {
                'name': 'Runtime state',
                'value': volume.runtime_state,
            },
        ]

    def get_filtered_field_names(self):
        return 'service_project_link',

    class Meta:
        model = models.Volume
        model_fields = ('name',)
        fields = ('service_project_link', 'backend_id', 'type', 'extra',) + model_fields
        read_only_fields = model_fields + ('backend_id', 'type', 'extra')


class VolumeImportSerializer(VolumeImportableSerializer):
    class Meta(VolumeImportableSerializer.Meta):
        fields = VolumeImportableSerializer.Meta.fields + ('url', 'uuid', 'created')
        read_only_fields = VolumeImportableSerializer.Meta.model_fields
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    @transaction.atomic
    def create(self, validated_data):
        service_project_link = validated_data['service_project_link']
        backend_id = validated_data['backend_id']

        if models.Volume.objects.filter(
            service_project_link__service__settings=service_project_link.service.settings,
            backend_id=backend_id
        ).exists():
            raise serializers.ValidationError({
                'backend_id': _('Volume has been imported already.')
            })

        try:
            backend = service_project_link.get_backend()
            volume = backend.import_volume(backend_id, save=True, service_project_link=service_project_link)
        except OpenStackBackendError:
            raise serializers.ValidationError({
                'backend_id': _("Can't import volume with ID %s") % validated_data['backend_id']
            })

        return volume


class VolumeAvailabilityZoneSerializer(BaseAvailabilityZoneSerializer):
    class Meta(BaseAvailabilityZoneSerializer.Meta):
        model = models.VolumeAvailabilityZone


class VolumeSerializer(structure_serializers.BaseResourceSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='openstacktenant-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-spl-detail',
        queryset=models.OpenStackTenantServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    action_details = serializers.JSONField(read_only=True)
    metadata = serializers.JSONField(read_only=True)
    instance_name = serializers.SerializerMethodField()
    type_name = serializers.CharField(source='type.name', read_only=True)
    availability_zone_name = serializers.CharField(source='availability_zone.name', read_only=True)

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Volume
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'source_snapshot', 'size', 'bootable', 'metadata',
            'image', 'image_metadata', 'image_name', 'type', 'type_name', 'runtime_state',
            'availability_zone', 'availability_zone_name',
            'device', 'action', 'action_details', 'instance', 'instance_name',
        )
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'image_metadata', 'image_name', 'bootable', 'source_snapshot', 'runtime_state', 'device', 'metadata',
            'action', 'instance'
        )
        protected_fields = structure_serializers.BaseResourceSerializer.Meta.protected_fields + (
            'size', 'image', 'type', 'availability_zone',
        )
        extra_kwargs = dict(
            instance={'lookup_field': 'uuid', 'view_name': 'openstacktenant-instance-detail'},
            image={'lookup_field': 'uuid', 'view_name': 'openstacktenant-image-detail'},
            source_snapshot={'lookup_field': 'uuid', 'view_name': 'openstacktenant-snapshot-detail'},
            type={'lookup_field': 'uuid', 'view_name': 'openstacktenant-volume-type-detail'},
            availability_zone={'lookup_field': 'uuid', 'view_name': 'openstacktenant-volume-availability-zone-detail'},
            size={'required': False, 'allow_null': True},
            **structure_serializers.BaseResourceSerializer.Meta.extra_kwargs
        )

    def get_instance_name(self, volume):
        if volume.instance:
            return volume.instance.name

    def validate(self, attrs):
        attrs = super(VolumeSerializer, self).validate(attrs)

        if self.instance is None:
            # image validation
            image = attrs.get('image')
            spl = attrs['service_project_link']
            if image and image.settings != spl.service.settings:
                raise serializers.ValidationError({'image': _('Image must belong to the same service settings')})
            # snapshot & size validation
            size = attrs.get('size')
            snapshot = attrs.get('snapshot')
            if not size and not snapshot:
                raise serializers.ValidationError(_('Snapshot or size should be defined'))
            if size and snapshot:
                raise serializers.ValidationError(_('It is impossible to define both snapshot and size'))
            # image & size validation
            size = size or snapshot.size
            if image and image.min_disk > size:
                raise serializers.ValidationError({
                    'size': _('Volume size should be equal or greater than %s for selected image') % image.min_disk
                })
            # type validation
            type = attrs.get('type')
            if type and type.settings != spl.service.settings:
                raise serializers.ValidationError({'type': _('Volume type must belong to the same service settings')})

            availability_zone = attrs.get('availability_zone')
            if availability_zone and availability_zone.settings != spl.service.settings:
                raise serializers.ValidationError(
                    _('Availability zone must belong to the same service settings.'))
            if availability_zone and not availability_zone.available:
                raise serializers.ValidationError(_('Zone is not available.'))
            if not availability_zone and django_settings.WALDUR_OPENSTACK_TENANT['REQUIRE_AVAILABILITY_ZONE']:
                if models.VolumeAvailabilityZone.objects.filter(settings=spl.service.settings).count() > 0:
                    raise serializers.ValidationError(_('Availability zone is mandatory.'))

        return attrs

    def create(self, validated_data):
        if not validated_data.get('size'):
            validated_data['size'] = validated_data['snapshot'].size
        if validated_data.get('image'):
            validated_data['image_name'] = validated_data['image'].name
        return super(VolumeSerializer, self).create(validated_data)


class VolumeExtendSerializer(serializers.Serializer):
    disk_size = serializers.IntegerField(min_value=1, label='Disk size')

    def validate_disk_size(self, disk_size):
        if disk_size < self.instance.size + 1024:
            raise serializers.ValidationError(
                _('Disk size should be greater or equal to %s') % (self.instance.size + 1024))
        return disk_size

    @transaction.atomic
    def update(self, instance, validated_data):
        new_size = validated_data.get('disk_size')

        settings = instance.service_project_link.service.settings
        spl = instance.service_project_link

        for quota_holder in [settings, spl]:
            quota_holder.add_quota_usage(quota_holder.Quotas.storage, new_size - instance.size, validate=True)

        instance.size = new_size
        instance.save(update_fields=['size'])
        return instance


class VolumeAttachSerializer(structure_serializers.PermissionFieldFilteringMixin,
                             serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.Volume
        fields = ['instance']
        extra_kwargs = dict(
            instance={
                'required': True,
                'allow_null': False,
                'view_name': 'openstacktenant-instance-detail',
                'lookup_field': 'uuid',
            }
        )

    def get_fields(self):
        fields = super(VolumeAttachSerializer, self).get_fields()
        volume = self.instance
        if volume:
            fields['instance'].display_name_field = 'name'
            fields['instance'].query_params = {
                'project_uuid': volume.service_project_link.project.uuid.hex,
                'service_uuid': volume.service_project_link.service.uuid.hex,
            }
        return fields

    def get_filtered_field_names(self):
        return ('instance',)

    def validate_instance(self, instance):
        States, RuntimeStates = models.Instance.States, models.Instance.RuntimeStates
        if instance.state != States.OK or instance.runtime_state not in (RuntimeStates.SHUTOFF, RuntimeStates.ACTIVE):
            raise serializers.ValidationError(
                _('Volume can be attached only to shutoff or active instance in OK state.'))
        volume = self.instance
        if instance.service_project_link != volume.service_project_link:
            raise serializers.ValidationError(_('Volume and instance should belong to the same service and project.'))
        if volume.availability_zone and instance.availability_zone:
            valid_zones = get_valid_availability_zones(volume)
            if valid_zones and valid_zones.get(instance.availability_zone.name) != volume.availability_zone.name:
                raise serializers.ValidationError(
                    _('Volume cannot be attached to virtual machine related to the other availability zone.'))
        return instance


class VolumeRetypeSerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model = models.Volume
        fields = ['type']

    type = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-volume-type-detail',
        queryset=models.VolumeType.objects.all(),
        lookup_field='uuid',
        allow_null=False,
        required=True,
    )

    def validate_type(self, type):
        volume = self.instance
        if type.settings != volume.service_project_link.service.settings:
            raise serializers.ValidationError(_('Volume and type should belong to the same service.'))
        if type == volume.type:
            raise serializers.ValidationError(_('Volume already has requested type.'))
        return type


class SnapshotRestorationSerializer(core_serializers.AugmentedSerializerMixin, serializers.HyperlinkedModelSerializer):
    name = serializers.CharField(write_only=True, help_text=_('New volume name.'))
    description = serializers.CharField(required=False, help_text=_('New volume description.'))
    volume_state = serializers.CharField(source='volume.human_readable_state', read_only=True)

    class Meta:
        model = models.SnapshotRestoration
        fields = ('uuid', 'created', 'name', 'description',
                  'volume', 'volume_name', 'volume_state', 'volume_runtime_state', 'volume_size', 'volume_device')
        read_only_fields = ('uuid', 'created', 'volume')
        related_paths = {
            'volume': ('name', 'state', 'runtime_state', 'size', 'device')
        }
        extra_kwargs = dict(
            volume={'lookup_field': 'uuid', 'view_name': 'openstacktenant-volume-detail'},
        )

    @transaction.atomic
    def create(self, validated_data):
        snapshot = self.context['view'].get_object()
        validated_data['snapshot'] = snapshot
        description = validated_data.pop('description', None) or 'Restored from snapshot %s' % snapshot.name

        volume = models.Volume(
            source_snapshot=snapshot,
            service_project_link=snapshot.service_project_link,
            name=validated_data.pop('name'),
            description=description,
            size=snapshot.size,
        )

        volume.save()
        volume.increase_backend_quotas_usage()
        validated_data['volume'] = volume

        return super(SnapshotRestorationSerializer, self).create(validated_data)


class SnapshotSerializer(structure_serializers.BaseResourceActionSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='openstacktenant-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-spl-detail',
        read_only=True)

    source_volume_name = serializers.ReadOnlyField(source='source_volume.name')
    action_details = serializers.JSONField(read_only=True)
    metadata = serializers.JSONField(required=False)
    restorations = SnapshotRestorationSerializer(many=True, read_only=True)
    snapshot_schedule_uuid = serializers.ReadOnlyField(source='snapshot_schedule.uuid')

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Snapshot
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'source_volume', 'size', 'metadata', 'runtime_state', 'source_volume_name', 'action', 'action_details',
            'restorations', 'kept_until', 'snapshot_schedule', 'snapshot_schedule_uuid'
        )
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'size', 'source_volume', 'metadata', 'runtime_state', 'action', 'snapshot_schedule',
            'service_settings', 'project',
        )
        extra_kwargs = dict(
            source_volume={'lookup_field': 'uuid', 'view_name': 'openstacktenant-volume-detail'},
            snapshot_schedule={'lookup_field': 'uuid', 'view_name': 'openstacktenant-snapshot-schedule-detail'},
            **structure_serializers.BaseResourceSerializer.Meta.extra_kwargs
        )

    def validate(self, attrs):
        # Skip validation on update
        if self.instance:
            return attrs

        attrs['source_volume'] = source_volume = self.context['view'].get_object()
        attrs['service_project_link'] = source_volume.service_project_link
        attrs['size'] = source_volume.size
        return super(SnapshotSerializer, self).validate(attrs)


# XXX: Consider either dropping or adapting to VolumeSerializer.
# XXX: Unclear when import can be triggered from HomePort
class SnapshotImportableSerializer(core_serializers.AugmentedSerializerMixin,
                                   serializers.HyperlinkedModelSerializer):
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-spl-detail',
        queryset=models.OpenStackTenantServiceProjectLink.objects.all(),
        write_only=True)
    source_volume_name = serializers.ReadOnlyField(source='source_volume.name')
    type = serializers.SerializerMethodField()

    def get_type(self, obj):
        return SupportedServices.get_name_for_model(models.Instance)

    def get_filtered_field_names(self):
        return 'service_project_link',

    class Meta:
        model = models.Snapshot
        model_fields = ('name', 'description', 'size', 'action', 'action_details',
                        'metadata', 'runtime_state', 'state', 'source_volume_name', 'source_volume_name')
        fields = ('service_project_link', 'backend_id', 'type') + model_fields
        read_only_fields = model_fields + ('backend_id', 'type')
        extra_kwargs = dict(
            source_volume={'lookup_field': 'uuid', 'view_name': 'openstacktenant-volume-detail'},
        )


class SnapshotImportSerializer(SnapshotImportableSerializer):
    class Meta(SnapshotImportableSerializer.Meta):
        fields = SnapshotImportableSerializer.Meta.fields + ('url', 'uuid', 'created')
        read_only_fields = SnapshotImportableSerializer.Meta.model_fields
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    @transaction.atomic
    def create(self, validated_data):
        service_project_link = validated_data['service_project_link']
        backend_id = validated_data['backend_id']

        if models.Snapshot.objects.filter(
            service_project_link__service__settings=service_project_link.service.settings,
            backend_id=backend_id
        ).exists():
            raise serializers.ValidationError({
                'backend_id': _('Snapshot has been imported already.')
            })

        try:
            backend = service_project_link.get_backend()
            snapshot = backend.import_snapshot(backend_id, save=True, service_project_link=service_project_link)
        except OpenStackBackendError:
            raise serializers.ValidationError({
                'backend_id': _("Can't import snapshot with ID %s") % validated_data['backend_id']
            })

        return snapshot


class NestedVolumeSerializer(core_serializers.AugmentedSerializerMixin,
                             serializers.HyperlinkedModelSerializer,
                             structure_serializers.BasicResourceSerializer):
    state = serializers.ReadOnlyField(source='get_state_display')
    type_name = serializers.CharField(source='type.name', read_only=True)

    class Meta:
        model = models.Volume
        fields = ('url', 'uuid', 'name', 'image_name', 'state', 'bootable', 'size', 'device', 'resource_type',
                  'type', 'type_name')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'type': {'lookup_field': 'uuid', 'view_name': 'openstacktenant-volume-type-detail'},
        }


class NestedSecurityGroupRuleSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.SecurityGroupRule
        fields = ('id', 'protocol', 'from_port', 'to_port', 'cidr')

    def to_internal_value(self, data):
        # Return exist security group as internal value if id is provided
        if 'id' in data:
            try:
                return models.SecurityGroupRule.objects.get(id=data['id'])
            except models.SecurityGroup:
                raise serializers.ValidationError(_('Security group with id %s does not exist') % data['id'])
        else:
            internal_data = super(NestedSecurityGroupRuleSerializer, self).to_internal_value(data)
            return models.SecurityGroupRule(**internal_data)


class NestedSecurityGroupSerializer(core_serializers.AugmentedSerializerMixin,
                                    core_serializers.HyperlinkedRelatedModelSerializer):
    rules = NestedSecurityGroupRuleSerializer(
        many=True,
        read_only=True,
    )
    state = serializers.ReadOnlyField(source='human_readable_state')

    class Meta:
        model = models.SecurityGroup
        fields = ('url', 'name', 'rules', 'description', 'state')
        read_only_fields = ('name', 'rules', 'description', 'state')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'}
        }


class NestedInternalIPSerializer(core_serializers.AugmentedSerializerMixin, serializers.HyperlinkedModelSerializer):

    class Meta:
        model = models.InternalIP
        fields = (
            'ip4_address', 'mac_address', 'subnet', 'subnet_uuid', 'subnet_name', 'subnet_description', 'subnet_cidr')
        read_only_fields = (
            'ip4_address', 'mac_address', 'subnet_uuid', 'subnet_name', 'subnet_description', 'subnet_cidr')
        related_paths = {
            'subnet': ('uuid', 'name', 'description', 'cidr'),
        }
        extra_kwargs = {
            'subnet': {'lookup_field': 'uuid', 'view_name': 'openstacktenant-subnet-detail'},
        }

    def to_internal_value(self, data):
        internal_value = super(NestedInternalIPSerializer, self).to_internal_value(data)
        return models.InternalIP(subnet=internal_value['subnet'], settings=internal_value['subnet'].settings)


class NestedFloatingIPSerializer(core_serializers.AugmentedSerializerMixin,
                                 core_serializers.HyperlinkedRelatedModelSerializer):
    subnet = serializers.HyperlinkedRelatedField(
        queryset=models.SubNet.objects.all(),
        source='internal_ip.subnet',
        view_name='openstacktenant-subnet-detail',
        lookup_field='uuid')
    subnet_uuid = serializers.ReadOnlyField(source='internal_ip.subnet.uuid')
    subnet_name = serializers.ReadOnlyField(source='internal_ip.subnet.name')
    subnet_description = serializers.ReadOnlyField(source='internal_ip.subnet.description')
    subnet_cidr = serializers.ReadOnlyField(source='internal_ip.subnet.cidr')

    class Meta:
        model = models.FloatingIP
        fields = ('url', 'uuid', 'address', 'internal_ip_ip4_address', 'internal_ip_mac_address',
                  'subnet', 'subnet_uuid', 'subnet_name', 'subnet_description', 'subnet_cidr')
        read_only_fields = ('address', 'internal_ip_ip4_address', 'internal_ip_mac_address')
        related_paths = {
            'internal_ip': ('ip4_address', 'mac_address')
        }
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'openstacktenant-fip-detail'},
        }

    def to_internal_value(self, data):
        """
        Return pair (floating_ip, subnet) as internal value.

        On floating IP creation user should specify what subnet should be used
        for connection and may specify what exactly floating IP should be used.
        If floating IP is not specified it will be represented as None.
        """
        floating_ip = None
        if 'url' in data:
            # use HyperlinkedRelatedModelSerializer (parent of NestedFloatingIPSerializer)
            # method to convert "url" to FloatingIP object
            floating_ip = super(NestedFloatingIPSerializer, self).to_internal_value(data)

        # use HyperlinkedModelSerializer (parent of HyperlinkedRelatedModelSerializer)
        # to convert "subnet" to SubNet object
        internal_value = super(core_serializers.HyperlinkedRelatedModelSerializer, self).to_internal_value(data)
        subnet = internal_value['internal_ip']['subnet']

        return floating_ip, subnet


def _validate_instance_internal_ips(internal_ips, settings):
    """ - make sure that internal_ips belong to specified setting;
        - make sure that internal_ips does not connect to the same subnet twice;
    """
    if not internal_ips:
        raise serializers.ValidationError(
            {'internal_ips_set': _('Instance should be connected to at least one network.')})
    subnets = [internal_ip.subnet for internal_ip in internal_ips]
    for subnet in subnets:
        if subnet.settings != settings:
            message = _('Subnet %s does not belong to the same service settings as service project link.') % subnet
            raise serializers.ValidationError({'internal_ips_set': message})
    pairs = [(internal_ip.subnet, internal_ip.backend_id) for internal_ip in internal_ips]
    duplicates = [subnet for subnet, count in collections.Counter(pairs).items() if count > 1]
    if duplicates:
        raise serializers.ValidationError(_('It is impossible to connect to subnet %s twice.') % duplicates[0][0])


def _validate_instance_security_groups(security_groups, settings):
    """ Make sure that security_group belong to specified setting.
    """
    for security_group in security_groups:
        if security_group.settings != settings:
            error = _('Security group %s does not belong to the same service settings as service project link.')
            raise serializers.ValidationError({'security_groups': error % security_group.name})


def _validate_instance_floating_ips(floating_ips_with_subnets, settings, instance_subnets):
    if floating_ips_with_subnets and 'external_network_id' not in settings.options:
        raise serializers.ValidationError(
            ugettext('Please specify tenant external network to perform floating IP operations.'))

    for floating_ip, subnet in floating_ips_with_subnets:
        if subnet not in instance_subnets:
            message = ugettext('SubNet %s is not connected to instance.') % subnet
            raise serializers.ValidationError({'floating_ips': message})
        if not floating_ip:
            continue
        if floating_ip.is_booked:
            message = ugettext('Floating IP %s is already booked for another instance creation')
            raise serializers.ValidationError({'floating_ips': message % floating_ip})
        if floating_ip.settings != settings:
            message = ugettext('Floating IP %s does not belong to the same service settings as service project link.')
            raise serializers.ValidationError({'floating_ips': message % floating_ip})

    subnets = [subnet for _, subnet in floating_ips_with_subnets]
    duplicates = [subnet for subnet, count in collections.Counter(subnets).items() if count > 1]
    if duplicates:
        raise serializers.ValidationError(ugettext('It is impossible to use subnet %s twice.') % duplicates[0])


def _validate_instance_name(data, max_len=255):
    """Copy paste from https://github.com/openstack/neutron-lib/blob/master/neutron_lib/api/validators/dns.py#L23"""

    # allow data to be lowercase. Internally OpenStack allows more flexibility
    # with hostnames as sanitizing happens, but we are more strict and want to preserve name <-> hostname mapping
    # https://github.com/openstack/nova/blob/e80300ac20388890539a7f709e526a0a5ba8e63d/nova/utils.py#L388

    DNS_LABEL_REGEX = '^([a-zA-Z0-9-]{1,63})$'
    try:
        # A trailing period is allowed to indicate that a name is fully
        # qualified per RFC 1034 (page 7).
        trimmed = data[:-1] if data.endswith('.') else data
        if len(trimmed) > max_len:
            raise TypeError(
                _("'%(trimmed)s' exceeds the %(maxlen)s character FQDN "
                  "limit") % {'trimmed': trimmed, 'maxlen': max_len})
        labels = trimmed.split('.')
        for label in labels:
            if not label:
                raise TypeError(_("Encountered an empty component"))
            if label.endswith('-') or label.startswith('-'):
                raise TypeError(
                    _("Name '%s' must not start or end with a hyphen") % label)
            if not re.match(DNS_LABEL_REGEX, label):
                raise TypeError(
                    _("Name '%s' must be 1-63 characters long, each of "
                      "which can only be alphanumeric or a hyphen") % label)
        # RFC 1123 hints that a TLD can't be all numeric. last is a TLD if
        # it's an FQDN.
        if len(labels) > 1 and re.match("^[0-9]+$", labels[-1]):
            raise TypeError(
                _("TLD '%s' must not be all numeric") % labels[-1])
    except TypeError as e:
        msg = _("'%(data)s' not a valid PQDN or FQDN. Reason: %(reason)s") % {
            'data': data, 'reason': e}
        raise serializers.ValidationError({'name': msg})


def _connect_floating_ip_to_instance(floating_ip, subnet, instance):
    """ Connect floating IP to instance via specified subnet.
        If floating IP is not defined - take exist free one or create a new one.
    """
    settings = instance.service_project_link.service.settings
    external_network_id = settings.options.get('external_network_id')
    if not core_utils.is_uuid_like(external_network_id):
        raise serializers.ValidationError(
            ugettext('Service provider does not have valid value of external_network_id'))

    if not floating_ip:
        kwargs = {
            'settings': settings,
            'is_booked': False,
            'backend_network_id': external_network_id,
        }
        # TODO: figure out why internal_ip__isnull throws errors when added to kwargs
        floating_ip = models.FloatingIP.objects.filter(internal_ip__isnull=True).filter(**kwargs).first()
        if not floating_ip:
            floating_ip = models.FloatingIP(**kwargs)
            floating_ip.increase_backend_quotas_usage()
    floating_ip.is_booked = True
    floating_ip.internal_ip = models.InternalIP.objects.filter(instance=instance, subnet=subnet).first()
    floating_ip.save()
    return floating_ip


class InstanceAvailabilityZoneSerializer(BaseAvailabilityZoneSerializer):
    class Meta(BaseAvailabilityZoneSerializer.Meta):
        model = models.InstanceAvailabilityZone


class DataVolumeSerializer(serializers.Serializer):
    size = serializers.IntegerField()
    volume_type = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-volume-type-detail',
        queryset=models.VolumeType.objects.all(),
        lookup_field='uuid',
        allow_null=True,
        required=False,
    )


class InstanceSerializer(structure_serializers.VirtualMachineSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='openstacktenant-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-spl-detail',
        queryset=models.OpenStackTenantServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    flavor = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-flavor-detail',
        lookup_field='uuid',
        queryset=models.Flavor.objects.all().select_related('settings'),
        write_only=True)

    image = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-image-detail',
        lookup_field='uuid',
        queryset=models.Image.objects.all().select_related('settings'),
        write_only=True)

    security_groups = NestedSecurityGroupSerializer(
        queryset=models.SecurityGroup.objects.all(), many=True, required=False)
    internal_ips_set = NestedInternalIPSerializer(many=True, required=False)
    floating_ips = NestedFloatingIPSerializer(
        queryset=models.FloatingIP.objects.all().filter(internal_ip__isnull=True),
        many=True,
        required=False
    )

    system_volume_size = serializers.IntegerField(min_value=1024, write_only=True)
    system_volume_type = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-volume-type-detail',
        queryset=models.VolumeType.objects.all(),
        lookup_field='uuid',
        allow_null=True,
        required=False,
        write_only=True,
    )
    data_volume_size = serializers.IntegerField(min_value=1024, required=False, write_only=True)
    data_volume_type = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-volume-type-detail',
        queryset=models.VolumeType.objects.all(),
        lookup_field='uuid',
        allow_null=True,
        required=False,
        write_only=True,
    )
    data_volumes = DataVolumeSerializer(many=True, required=False, write_only=True)
    volumes = NestedVolumeSerializer(many=True, required=False, read_only=True)
    action_details = serializers.JSONField(read_only=True)

    availability_zone_name = serializers.CharField(source='availability_zone.name', read_only=True)
    tenant_uuid = serializers.SerializerMethodField()

    class Meta(structure_serializers.VirtualMachineSerializer.Meta):
        model = models.Instance
        fields = structure_serializers.VirtualMachineSerializer.Meta.fields + (
            'image', 'flavor', 'flavor_disk', 'flavor_name',
            'system_volume_size', 'system_volume_type', 'data_volume_size',
            'data_volume_type', 'volumes', 'data_volumes',
            'security_groups', 'internal_ips', 'floating_ips', 'internal_ips_set',
            'availability_zone', 'availability_zone_name',
            'runtime_state', 'action', 'action_details', 'tenant_uuid',
        )
        protected_fields = structure_serializers.VirtualMachineSerializer.Meta.protected_fields + (
            'flavor', 'image', 'system_volume_size', 'data_volume_size',
            'floating_ips', 'security_groups', 'internal_ips_set',
            'availability_zone',
        )
        read_only_fields = structure_serializers.VirtualMachineSerializer.Meta.read_only_fields + (
            'flavor_disk', 'runtime_state', 'flavor_name', 'action',
        )
        extra_kwargs = dict(
            availability_zone={
                'lookup_field': 'uuid',
                'view_name': 'openstacktenant-instance-availability-zone-detail'
            },
            **structure_serializers.VirtualMachineSerializer.Meta.extra_kwargs
        )

    def get_fields(self):
        fields = super(InstanceSerializer, self).get_fields()
        floating_ip_field = fields.get('floating_ips')
        if floating_ip_field:
            floating_ip_field.value_field = 'url'
            floating_ip_field.display_name_field = 'address'

        return fields

    def get_tenant_uuid(self, instance):
        service_settings = instance.service_project_link.service.settings
        tenant = service_settings.scope
        if not tenant:
            return
        if not isinstance(tenant, openstack_models.Tenant):
            return
        try:
            request = self.context['request']
            user = request.user
        except (KeyError, AttributeError):
            return
        project = tenant.service_project_link.project
        if not _has_admin_access(user, project):
            return
        return tenant.uuid.hex

    @staticmethod
    def eager_load(queryset, request):
        queryset = structure_serializers.VirtualMachineSerializer.eager_load(queryset, request)
        return queryset.prefetch_related(
            'security_groups',
            'security_groups__rules',
            'volumes',
        )

    def validate(self, attrs):
        attrs = super(InstanceSerializer, self).validate(attrs)

        # skip validation on object update
        if self.instance is not None:
            return attrs

        service_project_link = attrs['service_project_link']
        settings = service_project_link.service.settings
        flavor = attrs['flavor']
        image = attrs['image']

        if any([flavor.settings != settings, image.settings != settings]):
            raise serializers.ValidationError(
                _('Flavor and image must belong to the same service settings as service project link.'))

        if image.min_ram > flavor.ram:
            raise serializers.ValidationError(
                {'flavor': _('RAM of flavor is not enough for selected image %s') % image.min_ram})

        if image.min_disk > attrs['system_volume_size']:
            raise serializers.ValidationError(
                {'system_volume_size': _('System volume size has to be greater than %s') % image.min_disk})

        internal_ips = attrs.get('internal_ips_set', [])
        _validate_instance_security_groups(attrs.get('security_groups', []), settings)
        _validate_instance_internal_ips(internal_ips, settings)
        subnets = [internal_ip.subnet for internal_ip in internal_ips]
        _validate_instance_floating_ips(attrs.get('floating_ips', []), settings, subnets)
        _validate_instance_name(attrs.get('name'))

        availability_zone = attrs.get('availability_zone')
        if availability_zone and availability_zone.settings != settings:
            raise serializers.ValidationError(
                _('Instance and availability zone must belong to the same service settings as service project link.'))
        if availability_zone and not availability_zone.available:
            raise serializers.ValidationError(_('Zone is not available.'))

        if not availability_zone and django_settings.WALDUR_OPENSTACK_TENANT['REQUIRE_AVAILABILITY_ZONE']:
            if models.InstanceAvailabilityZone.objects.filter(settings=settings).count() > 0:
                raise serializers.ValidationError(_('Availability zone is mandatory.'))

        return attrs

    def _find_volume_availability_zone(self, instance):
        # Find volume AZ using instance AZ. It is assumed that user can't select arbitrary
        # combination of volume and instance AZ. Once instance AZ is selected,
        # volume AZ is taken from settings.

        volume_availability_zone = None
        valid_zones = get_valid_availability_zones(instance)
        if instance.availability_zone and valid_zones:
            volume_availability_zone_name = valid_zones.get(instance.availability_zone.name)
            if volume_availability_zone_name:
                try:
                    volume_availability_zone = models.VolumeAvailabilityZone.objects.get(
                        name=volume_availability_zone_name,
                        settings=instance.service_project_link.service.settings,
                        available=True,
                    )
                except models.VolumeAvailabilityZone.DoesNotExist:
                    pass
        return volume_availability_zone

    @transaction.atomic
    def create(self, validated_data):
        """ Store flavor, ssh_key and image details into instance model.
            Create volumes and security groups for instance.
        """
        security_groups = validated_data.pop('security_groups', [])
        internal_ips = validated_data.pop('internal_ips_set', [])
        floating_ips_with_subnets = validated_data.pop('floating_ips', [])
        spl = validated_data['service_project_link']
        ssh_key = validated_data.get('ssh_public_key')
        if ssh_key:
            # We want names to be human readable in backend.
            # OpenStack only allows latin letters, digits, dashes, underscores and spaces
            # as key names, thus we mangle the original name.
            safe_name = re.sub(r'[^-a-zA-Z0-9 _]+', '_', ssh_key.name)[:17]
            validated_data['key_name'] = '{0}-{1}'.format(ssh_key.uuid.hex, safe_name)
            validated_data['key_fingerprint'] = ssh_key.fingerprint

        flavor = validated_data['flavor']
        validated_data['flavor_name'] = flavor.name
        validated_data['cores'] = flavor.cores
        validated_data['ram'] = flavor.ram
        validated_data['flavor_disk'] = flavor.disk

        image = validated_data['image']
        validated_data['image_name'] = image.name
        validated_data['min_disk'] = image.min_disk
        validated_data['min_ram'] = image.min_ram

        system_volume_size = validated_data['system_volume_size']
        data_volume_size = validated_data.get('data_volume_size', 0)
        total_disk = data_volume_size + system_volume_size

        data_volumes = validated_data.get('data_volumes', [])
        if data_volumes:
            total_disk += sum(volume['size'] for volume in data_volumes)

        validated_data['disk'] = total_disk

        instance = super(InstanceSerializer, self).create(validated_data)

        # security groups
        instance.security_groups.add(*security_groups)
        # internal IPs
        for internal_ip in internal_ips:
            internal_ip.instance = instance
            internal_ip.save()
        # floating IPs
        for floating_ip, subnet in floating_ips_with_subnets:
            _connect_floating_ip_to_instance(floating_ip, subnet, instance)

        volume_availability_zone = self._find_volume_availability_zone(instance)

        # volumes
        volumes = []
        system_volume = models.Volume.objects.create(
            name='{0}-system'.format(instance.name[:143]),  # volume name cannot be longer than 150 symbols
            service_project_link=spl,
            size=system_volume_size,
            image=image,
            image_name=image.name,
            bootable=True,
            availability_zone=volume_availability_zone,
            type=validated_data.get('system_volume_type'),
        )
        volumes.append(system_volume)

        if data_volume_size:
            data_volume = models.Volume.objects.create(
                name='{0}-data'.format(instance.name[:145]),  # volume name cannot be longer than 150 symbols
                service_project_link=spl,
                size=data_volume_size,
                availability_zone=volume_availability_zone,
                type=validated_data.get('data_volume_type'),
            )
            volumes.append(data_volume)

        for index, volume in enumerate(data_volumes):
            data_volume = models.Volume.objects.create(
                name='{0}-data-{1}'.format(instance.name[:140], index + 2),  # volume name cannot be longer than 150 symbols
                service_project_link=spl,
                size=volume['size'],
                availability_zone=volume_availability_zone,
                type=volume.get('volume_type'),
            )
            volumes.append(data_volume)

        for volume in volumes:
            volume.increase_backend_quotas_usage()

        instance.volumes.add(*volumes)
        return instance


class InstanceFlavorChangeSerializer(structure_serializers.PermissionFieldFilteringMixin, serializers.Serializer):
    flavor = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-flavor-detail',
        lookup_field='uuid',
        queryset=models.Flavor.objects.all(),
    )

    def get_fields(self):
        fields = super(InstanceFlavorChangeSerializer, self).get_fields()
        if self.instance:
            fields['flavor'].query_params = {
                'settings_uuid': self.instance.service_project_link.service.settings.uuid
            }
        return fields

    def get_filtered_field_names(self):
        return ('flavor',)

    def validate_flavor(self, value):
        if value is not None:
            spl = self.instance.service_project_link

            if value.name == self.instance.flavor_name:
                raise serializers.ValidationError(
                    _('New flavor is the same as current.'))

            if value.settings != spl.service.settings:
                raise serializers.ValidationError(
                    _('New flavor is not within the same service settings'))

            if value.disk < self.instance.flavor_disk:
                raise serializers.ValidationError(
                    _('New flavor disk should be greater than the previous value'))
        return value

    @transaction.atomic
    def update(self, instance, validated_data):
        flavor = validated_data.get('flavor')

        spl = instance.service_project_link
        settings = spl.service.settings
        quota_holders = [spl, settings]

        # Service settings has optional field for related tenant resource.
        # We should update tenant quotas if related tenant is defined.
        # Otherwise stale quotas would be used for quota validation during instance provisioning.
        # Note that all tenant quotas are injected to service settings when application is bootstrapped.
        if settings.scope:
            quota_holders.append(settings.scope)

        for quota_holder in quota_holders:
            quota_holder.add_quota_usage(quota_holder.Quotas.ram, flavor.ram - instance.ram, validate=True)
            quota_holder.add_quota_usage(quota_holder.Quotas.vcpu, flavor.cores - instance.cores, validate=True)

        instance.ram = flavor.ram
        instance.cores = flavor.cores
        instance.flavor_disk = flavor.disk
        instance.flavor_name = flavor.name
        instance.save(update_fields=['ram', 'cores', 'flavor_name', 'flavor_disk'])
        return instance


class InstanceDeleteSerializer(serializers.Serializer):
    delete_volumes = serializers.BooleanField(default=True)
    release_floating_ips = serializers.BooleanField(label=_('Release floating IPs'), default=True)

    def validate(self, attrs):
        if attrs['delete_volumes'] and models.Snapshot.objects.filter(source_volume__instance=self.instance).exists():
            raise serializers.ValidationError(_('Cannot delete instance. One of its volumes has attached snapshot.'))
        return attrs


class InstanceSecurityGroupsUpdateSerializer(serializers.Serializer):
    security_groups = NestedSecurityGroupSerializer(
        queryset=models.SecurityGroup.objects.all(),
        many=True,
    )

    def get_fields(self):
        fields = super(InstanceSecurityGroupsUpdateSerializer, self).get_fields()
        instance = self.instance
        if instance:
            fields['security_groups'].display_name_field = 'name'
            fields['security_groups'].view_name = 'openstacktenant-sgp-detail'
            fields['security_groups'].query_params = {
                'settings_uuid': instance.service_project_link.service.settings.uuid
            }
        return fields

    def validate_security_groups(self, security_groups):
        spl = self.instance.service_project_link

        for security_group in security_groups:
            if security_group.settings != spl.service.settings:
                raise serializers.ValidationError(
                    _('Security group %s is not within the same service settings') % security_group.name)

        return security_groups

    @transaction.atomic
    def update(self, instance, validated_data):
        security_groups = validated_data.pop('security_groups', None)
        if security_groups is not None:
            instance.security_groups.clear()
            instance.security_groups.add(*security_groups)

        return instance


class InstanceInternalIPsSetUpdateSerializer(serializers.Serializer):
    internal_ips_set = NestedInternalIPSerializer(many=True)

    def get_fields(self):
        fields = super(InstanceInternalIPsSetUpdateSerializer, self).get_fields()
        instance = self.instance
        if instance:
            fields['internal_ips_set'].view_name = 'openstacktenant-subnet-detail'
            fields['internal_ips_set'].query_params = {
                'settings_uuid': instance.service_project_link.service.settings.uuid
            }
        return fields

    def validate_internal_ips_set(self, internal_ips_set):
        spl = self.instance.service_project_link
        _validate_instance_internal_ips(internal_ips_set, spl.service.settings)
        return internal_ips_set

    @transaction.atomic
    def update(self, instance, validated_data):
        internal_ips_set = validated_data['internal_ips_set']
        new_subnets = [ip.subnet for ip in internal_ips_set]
        # delete stale IPs
        models.InternalIP.objects.filter(instance=instance).exclude(subnet__in=new_subnets).delete()
        # create new IPs
        for internal_ip in internal_ips_set:
            match = models.InternalIP.objects.filter(instance=instance, subnet=internal_ip.subnet).first()
            if not match:
                models.InternalIP.objects.create(instance=instance,
                                                 subnet=internal_ip.subnet,
                                                 settings=internal_ip.subnet.settings)

        return instance


class InstanceFloatingIPsUpdateSerializer(serializers.Serializer):
    floating_ips = NestedFloatingIPSerializer(queryset=models.FloatingIP.objects.all(), many=True, required=False)

    def get_fields(self):
        fields = super(InstanceFloatingIPsUpdateSerializer, self).get_fields()
        instance = self.instance
        if instance:
            queryset = models.FloatingIP.objects.all().filter(
                Q(internal_ip__isnull=True) | Q(internal_ip__instance=instance)
            )
            fields['floating_ips'] = NestedFloatingIPSerializer(queryset=queryset, many=True, required=False)
            fields['floating_ips'].view_name = 'openstacktenant-fip-detail'
            fields['floating_ips'].query_params = {
                'settings_uuid': instance.service_project_link.service.settings.uuid.hex,
                'is_booked': False,
                'free': True,
            }
        return fields

    def validate_floating_ips(self, floating_ips):
        spl = self.instance.service_project_link
        subnets = self.instance.subnets.all()
        _validate_instance_floating_ips(floating_ips, spl.service.settings, subnets)
        return floating_ips

    def update(self, instance, validated_data):
        floating_ips_with_subnets = validated_data['floating_ips']
        floating_ips_to_disconnect = list(self.instance.floating_ips)

        # Store both old and new floating IP addresses for action event logger
        new_floating_ips = [floating_ip for (floating_ip, subnet) in floating_ips_with_subnets if floating_ip]
        instance._old_floating_ips = [floating_ip.address for floating_ip in floating_ips_to_disconnect]
        instance._new_floating_ips = [floating_ip.address for floating_ip in new_floating_ips]

        for floating_ip, subnet in floating_ips_with_subnets:
            if floating_ip in floating_ips_to_disconnect:
                floating_ips_to_disconnect.remove(floating_ip)
                continue
            _connect_floating_ip_to_instance(floating_ip, subnet, instance)
        for floating_ip in floating_ips_to_disconnect:
            floating_ip.internal_ip = None
            floating_ip.save()
        return instance


class BackupRestorationSerializer(serializers.HyperlinkedModelSerializer):
    name = serializers.CharField(
        required=False, help_text=_('New instance name. Leave blank to use source instance name.'))
    security_groups = NestedSecurityGroupSerializer(
        queryset=models.SecurityGroup.objects.all(), many=True, required=False)
    internal_ips_set = NestedInternalIPSerializer(many=True, required=False)
    floating_ips = NestedFloatingIPSerializer(
        queryset=models.FloatingIP.objects.all().filter(internal_ip__isnull=True),
        many=True,
        required=False
    )

    class Meta:
        model = models.BackupRestoration
        fields = ('uuid', 'instance', 'created', 'flavor', 'name', 'floating_ips', 'security_groups',
                  'internal_ips_set')
        read_only_fields = ('url', 'uuid', 'instance', 'created', 'backup')
        extra_kwargs = dict(
            instance={'lookup_field': 'uuid', 'view_name': 'openstacktenant-instance-detail'},
            flavor={'lookup_field': 'uuid', 'view_name': 'openstacktenant-flavor-detail', 'allow_null': False,
                    'required': True},
        )

    def get_fields(self):
        fields = super(BackupRestorationSerializer, self).get_fields()
        view = self.context.get('view')  # On docs generation context does not contain "view".
        if view and view.action == 'restore' and self.instance:
            backup = self.instance
            settings = backup.instance.service_project_link.service.settings
            fields['flavor'].display_name_field = 'name'
            fields['flavor'].view_name = 'openstacktenant-flavor-detail'
            fields['flavor'].query_params = {
                'settings_uuid': backup.service_project_link.service.settings.uuid.hex,
            }

            floating_ip_field = fields.get('floating_ips')
            if floating_ip_field:
                floating_ip_field.view_name = 'openstacktenant-fip-detail'
                floating_ip_field.query_params = {
                    'settings_uuid': settings.uuid.hex,
                    'is_booked': False,
                    'free': True,
                }
                floating_ip_field.display_name_field = 'address'

            internal_ips_set_field = fields.get('internal_ips_set')
            if internal_ips_set_field:
                internal_ips_set_field.query_params = {
                    'settings_uuid': settings.uuid.hex,
                }
                internal_ips_set_field.view_name = 'openstacktenant-subnet-detail'
                internal_ips_set_field.display_name_field = 'name'

            security_groups_field = fields.get('security_groups')
            if security_groups_field:
                security_groups_field.query_params = {
                    'settings_uuid': settings.uuid.hex,
                }
                security_groups_field.view_name = 'openstacktenant-sgp-detail'
                security_groups_field.display_name_field = 'name'

        return fields

    def validate(self, attrs):
        flavor = attrs['flavor']
        backup = self.context['view'].get_object()
        try:
            backup.instance.volumes.get(bootable=True)
        except ObjectDoesNotExist:
            raise serializers.ValidationError(_('OpenStack instance should have bootable volume.'))

        settings = backup.instance.service_project_link.service.settings

        if flavor.settings != settings:
            raise serializers.ValidationError({'flavor': _('Flavor is not within services\' settings.')})

        _validate_instance_security_groups(attrs.get('security_groups', []), settings)

        internal_ips = attrs.get('internal_ips_set', [])
        _validate_instance_internal_ips(internal_ips, settings)

        subnets = [internal_ip.subnet for internal_ip in internal_ips]
        _validate_instance_floating_ips(attrs.get('floating_ips', []), settings, subnets)

        return attrs

    @transaction.atomic
    def update(self, backup_instance, validated_data):
        flavor = validated_data['flavor']
        validated_data['backup'] = backup = backup_instance
        source_instance = backup.instance
        # instance that will be restored
        metadata = backup.metadata or {}
        instance = models.Instance.objects.create(
            name=validated_data.pop('name', None) or metadata.get('name', source_instance.name),
            description=metadata.get('description', ''),
            service_project_link=backup.service_project_link,
            flavor_disk=flavor.disk,
            flavor_name=flavor.name,
            key_name=source_instance.key_name,
            key_fingerprint=source_instance.key_fingerprint,
            cores=flavor.cores,
            ram=flavor.ram,
            min_ram=metadata.get('min_ram', 0),
            min_disk=metadata.get('min_disk', 0),
            image_name=metadata.get('image_name', ''),
            user_data=metadata.get('user_data', ''),
            disk=sum([snapshot.size for snapshot in backup.snapshots.all()]),
        )

        instance.internal_ips_set.add(*validated_data.pop('internal_ips_set', []), bulk=False)
        instance.security_groups.add(*validated_data.pop('security_groups', []))

        for floating_ip, subnet in validated_data.pop('floating_ips', []):
            _connect_floating_ip_to_instance(floating_ip, subnet, instance)

        instance.increase_backend_quotas_usage()
        validated_data['instance'] = instance
        backup_restoration = super(BackupRestorationSerializer, self).create(validated_data)
        # restoration for each instance volume from snapshot.
        for snapshot in backup.snapshots.all():
            volume = models.Volume(
                source_snapshot=snapshot,
                service_project_link=snapshot.service_project_link,
                name='{0}-volume'.format(instance.name[:143]),
                description='Restored from backup %s' % backup.uuid.hex,
                size=snapshot.size,
            )
            volume.save()
            volume.increase_backend_quotas_usage()
            instance.volumes.add(volume)
        return backup_restoration


class BackupSerializer(structure_serializers.BaseResourceActionSerializer):
    # Serializer requires OpenStack Instance in context on creation
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='openstacktenant-detail',
        read_only=True,
        lookup_field='uuid')
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-spl-detail',
        read_only=True,
    )
    metadata = serializers.JSONField(read_only=True)
    instance_name = serializers.ReadOnlyField(source='instance.name')
    instance_security_groups = NestedSecurityGroupSerializer(
        read_only=True, many=True, source='instance.security_groups')
    instance_internal_ips_set = NestedInternalIPSerializer(
        read_only=True, many=True, source='instance.internal_ips_set')
    instance_floating_ips = NestedFloatingIPSerializer(
        read_only=True, many=True, source='instance.floating_ips')

    restorations = BackupRestorationSerializer(many=True, read_only=True)
    backup_schedule_uuid = serializers.ReadOnlyField(source='backup_schedule.uuid')

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Backup
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'kept_until', 'metadata', 'instance', 'instance_name', 'restorations',
            'backup_schedule', 'backup_schedule_uuid',
            'instance_security_groups', 'instance_internal_ips_set', 'instance_floating_ips')
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'instance', 'service_project_link', 'backup_schedule', 'service_settings', 'project')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'instance': {'lookup_field': 'uuid', 'view_name': 'openstacktenant-instance-detail'},
            'backup_schedule': {'lookup_field': 'uuid', 'view_name': 'openstacktenant-backup-schedule-detail'},
        }

    def validate(self, attrs):
        # Skip validation on update
        if self.instance:
            return attrs

        attrs['instance'] = instance = self.context['view'].get_object()
        attrs['service_project_link'] = instance.service_project_link
        attrs['metadata'] = self.get_backup_metadata(instance)
        return super(BackupSerializer, self).validate(attrs)

    @transaction.atomic
    def create(self, validated_data):
        backup = super(BackupSerializer, self).create(validated_data)
        self.create_backup_snapshots(backup)
        return backup

    @staticmethod
    def get_backup_metadata(instance):
        return {
            'name': instance.name,
            'description': instance.description,
            'min_ram': instance.min_ram,
            'min_disk': instance.min_disk,
            'size': instance.size,
            'key_name': instance.key_name,
            'key_fingerprint': instance.key_fingerprint,
            'user_data': instance.user_data,
            'flavor_name': instance.flavor_name,
            'image_name': instance.image_name,
        }

    @staticmethod
    def create_backup_snapshots(backup):
        for volume in backup.instance.volumes.all():
            snapshot = models.Snapshot.objects.create(
                name='Part of backup: %s (volume: %s)' % (backup.name[:60], volume.name[:60]),
                service_project_link=backup.service_project_link,
                size=volume.size,
                source_volume=volume,
                description='Part of backup %s (UUID: %s)' % (backup.name, backup.uuid.hex),
            )
            snapshot.increase_backend_quotas_usage()
            backup.snapshots.add(snapshot)


class BaseScheduleSerializer(structure_serializers.BaseResourceActionSerializer):
    timezone = serializers.ChoiceField(choices=[(t, t) for t in pytz.all_timezones],
                                       initial=timezone.get_current_timezone_name(),
                                       default=timezone.get_current_timezone_name())
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='openstacktenant-detail',
        read_only=True,
        lookup_field='uuid')
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-spl-detail',
        read_only=True,
    )

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'retention_time', 'timezone', 'maximal_number_of_resources', 'schedule',
            'is_active', 'next_trigger_at')
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'is_active', 'next_trigger_at', 'service_project_link', 'service_settings', 'project')


class BackupScheduleSerializer(BaseScheduleSerializer):

    class Meta(BaseScheduleSerializer.Meta):
        model = models.BackupSchedule
        fields = BaseScheduleSerializer.Meta.fields + (
            'instance', 'instance_name')
        read_only_fields = BaseScheduleSerializer.Meta.read_only_fields + (
            'backups', 'instance')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'instance': {'lookup_field': 'uuid', 'view_name': 'openstacktenant-instance-detail'},
        }
        related_paths = {
            'instance': ('name',),
        }

    def validate(self, attrs):
        # Skip validation on update
        if self.instance:
            return attrs

        instance = self.context['view'].get_object()
        if not instance.volumes.filter(bootable=True).exists():
            raise serializers.ValidationError(_('OpenStack instance should have bootable volume.'))
        attrs['instance'] = instance
        attrs['service_project_link'] = instance.service_project_link
        attrs['state'] = instance.States.OK
        return super(BackupScheduleSerializer, self).validate(attrs)


class SnapshotScheduleSerializer(BaseScheduleSerializer):

    class Meta(BaseScheduleSerializer.Meta):
        model = models.SnapshotSchedule
        fields = BaseScheduleSerializer.Meta.fields + (
            'source_volume', 'source_volume_name')
        read_only_fields = BaseScheduleSerializer.Meta.read_only_fields + (
            'snapshots', 'source_volume')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'source_volume': {'lookup_field': 'uuid', 'view_name': 'openstacktenant-volume-detail'},
        }
        related_paths = {
            'source_volume': ('name',),
        }

    def validate(self, attrs):
        # Skip validation on update
        if self.instance:
            return attrs

        volume = self.context['view'].get_object()
        attrs['source_volume'] = volume
        attrs['service_project_link'] = volume.service_project_link
        attrs['state'] = volume.States.OK
        return super(SnapshotScheduleSerializer, self).validate(attrs)


class MeterSampleSerializer(serializers.Serializer):
    name = serializers.CharField(source='counter_name')
    value = serializers.FloatField(source='counter_volume')
    type = serializers.CharField(source='counter_type')
    unit = serializers.CharField(source='counter_unit')
    timestamp = fields.StringTimestampField(formats=('%Y-%m-%dT%H:%M:%S.%f', '%Y-%m-%dT%H:%M:%S'))
    recorded_at = fields.StringTimestampField(formats=('%Y-%m-%dT%H:%M:%S.%f', '%Y-%m-%dT%H:%M:%S'))


class MeterTimestampIntervalSerializer(core_serializers.TimestampIntervalSerializer):
    def get_fields(self):
        fields = super(MeterTimestampIntervalSerializer, self).get_fields()
        fields['start'].default = core_utils.timeshift(hours=-1)
        fields['end'].default = core_utils.timeshift()
        return fields


def get_instance(openstack_floating_ip):
    # cache openstack instance on openstack floating_ip instance
    if hasattr(openstack_floating_ip, '_instance'):
        return openstack_floating_ip._instance
    if not openstack_floating_ip.backend_id or not openstack_floating_ip.address:
        openstack_floating_ip._instance = None
        return
    try:
        floating_ip = models.FloatingIP.objects\
            .exclude(internal_ip__isnull=True)\
            .get(backend_id=openstack_floating_ip.backend_id,
                 address=openstack_floating_ip.address)
    except models.FloatingIP.DoesNotExist:
        openstack_floating_ip._instance = None
    else:
        instance = getattr(floating_ip.internal_ip, 'instance', None)
        openstack_floating_ip._instance = instance
        return instance


def get_instance_attr(openstack_floating_ip, name):
    instance = get_instance(openstack_floating_ip)
    return getattr(instance, name, None)


def get_instance_uuid(serializer, openstack_floating_ip):
    return get_instance_attr(openstack_floating_ip, 'uuid')


def get_instance_name(serializer, openstack_floating_ip):
    return get_instance_attr(openstack_floating_ip, 'name')


def get_instance_url(serializer, openstack_floating_ip):
    instance = get_instance(openstack_floating_ip)
    if instance:
        return reverse('openstacktenant-instance-detail', kwargs={'uuid': instance.uuid.hex},
                       request=serializer.context['request'])


def add_instance_fields(sender, fields, **kwargs):
    fields['instance_uuid'] = serializers.SerializerMethodField()
    setattr(sender, 'get_instance_uuid', get_instance_uuid)
    fields['instance_name'] = serializers.SerializerMethodField()
    setattr(sender, 'get_instance_name', get_instance_name)
    fields['instance_url'] = serializers.SerializerMethodField()
    setattr(sender, 'get_instance_url', get_instance_url)


core_signals.pre_serializer_fields.connect(add_instance_fields, sender=openstack_serializers.FloatingIPSerializer)


class InstanceImportableSerializer(core_serializers.AugmentedSerializerMixin, serializers.HyperlinkedModelSerializer):
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-spl-detail',
        queryset=models.OpenStackTenantServiceProjectLink.objects.all(),
        write_only=True)
    type = serializers.SerializerMethodField()
    extra = serializers.SerializerMethodField()

    def get_extra(self, instance):
        return [
            {
                'name': 'Description',
                'value': instance.description,
            },
            {
                'name': 'Runtime state',
                'value': instance.runtime_state,
            },
            {
                'name': 'Flavor',
                'value': instance.flavor_name,
            },
            {
                'name': 'RAM (MBs)',
                'value': instance.ram,
            },
            {
                'name': 'Cores',
                'value': instance.cores,
            },

        ]

    def get_type(self, obj):
        return SupportedServices.get_name_for_model(models.Instance)

    def get_filtered_field_names(self):
        return 'service_project_link',

    class Meta:
        model = models.Instance
        model_fields = ('name',)
        fields = ('service_project_link', 'backend_id', 'type', 'extra') + model_fields
        read_only_fields = model_fields + ('backend_id', 'type', 'extra')


class InstanceImportSerializer(InstanceImportableSerializer):
    class Meta(InstanceImportableSerializer.Meta):
        fields = InstanceImportableSerializer.Meta.fields + ('url', 'uuid', 'created')
        read_only_fields = InstanceImportableSerializer.Meta.model_fields
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    @transaction.atomic
    def create(self, validated_data):
        service_project_link = validated_data['service_project_link']
        backend_id = validated_data['backend_id']

        if models.Instance.objects.filter(
            service_project_link__service__settings=service_project_link.service.settings,
            backend_id=backend_id
        ).exists():
            raise serializers.ValidationError({'backend_id': _('Instance has been imported already.')})

        try:
            backend = service_project_link.get_backend()
            instance = backend.import_instance(backend_id, save=True, service_project_link=service_project_link)
        except OpenStackBackendError:
            raise serializers.ValidationError({
                'backend_id': _("Can't import instance with ID %s") % validated_data['backend_id']
            })

        return instance


class ConsoleLogSerializer(serializers.Serializer):
    length = serializers.IntegerField(required=False)


class VolumeTypeSerializer(BaseVolumeTypeSerializer):
    class Meta(BaseVolumeTypeSerializer.Meta):
        model = models.VolumeType
        fields = BaseVolumeTypeSerializer.Meta.fields + ('is_default',)


class SharedSettingsCustomerSerializer(serializers.Serializer):
    name = serializers.ReadOnlyField()
    uuid = serializers.ReadOnlyField()
    created = serializers.ReadOnlyField()
    abbreviation = serializers.ReadOnlyField()
    vm_count = serializers.ReadOnlyField()



from django.apps import AppConfig


class AzureConfig(AppConfig):
    name = 'waldur_azure'
    verbose_name = 'Waldur Azure'
    service_name = 'Azure'
    is_public_service = True

    def ready(self):
        from waldur_core.structure import SupportedServices
        from .backend import AzureBackend
        SupportedServices.register_backend(AzureBackend)



import base64
from csv import DictWriter
from io import StringIO
import logging

from celery import shared_task, chain
from django.conf import settings
from django.db.models import Q
from django.template.loader import render_to_string
from django.utils import timezone

from waldur_core.core import utils as core_utils
from waldur_core.structure import models as structure_models
from waldur_mastermind.invoices.utils import get_previous_month

from . import models, registrators, serializers, utils


logger = logging.getLogger(__name__)


@shared_task(name='invoices.create_monthly_invoices')
def create_monthly_invoices():
    """
    - For every customer change state of the invoices for previous months from "pending" to "billed"
      and freeze their items.
    - Create new invoice for every customer in current month if not created yet.
    """
    date = timezone.now()

    old_invoices = models.Invoice.objects.filter(
        Q(state=models.Invoice.States.PENDING, year__lt=date.year) |
        Q(state=models.Invoice.States.PENDING, year=date.year, month__lt=date.month)
    )
    for invoice in old_invoices:
        invoice.set_created()

    customers = structure_models.Customer.objects.all()
    if settings.WALDUR_CORE['ENABLE_ACCOUNTING_START_DATE']:
        customers = customers.filter(accounting_start_date__lt=timezone.now())

    for customer in customers.iterator():
        registrators.RegistrationManager.get_or_create_invoice(customer, core_utils.month_start(date))

    if settings.WALDUR_INVOICES['INVOICE_REPORTING']['ENABLE']:
        send_invoice_report.delay()

    if settings.WALDUR_INVOICES['SEND_CUSTOMER_INVOICES']:
        chain(create_pdf_for_new_invoices.si(), send_new_invoices_notification.si())()
    else:
        create_pdf_for_new_invoices.delay()


@shared_task(name='invoices.send_invoice_notification')
def send_invoice_notification(invoice_uuid):
    """ Sends email notification with invoice link to customer owners """
    invoice = models.Invoice.objects.get(uuid=invoice_uuid)
    link_template = settings.WALDUR_INVOICES['INVOICE_LINK_TEMPLATE']

    if not link_template:
        logger.error('INVOICE_LINK_TEMPLATE is not set. '
                     'Sending of invoice notification is not available.')
        return

    if '{uuid}' not in link_template:
        logger.error('INVOICE_LINK_TEMPLATE must include \'{uuid}\' parameter. '
                     'Sending of invoice notification is not available.')
        return

    context = {
        'month': invoice.month,
        'year': invoice.year,
        'customer': invoice.customer.name,
        'link': link_template.format(uuid=invoice_uuid)
    }

    emails = [owner.email for owner in invoice.customer.get_owners()]
    filename = None
    attachment = None
    content_type = None

    if invoice._file:
        filename = '%s_%s_%s.pdf' % (settings.WALDUR_CORE['SITE_NAME'].replace(' ', '_'),
                                     invoice.year, invoice.month)
        attachment = base64.b64decode(invoice._file)
        content_type = 'application/pdf'

    logger.debug('About to send invoice {invoice} notification to {emails}'.format(invoice=invoice, emails=emails))
    core_utils.broadcast_mail('invoices', 'notification', context, emails,
                              filename=filename, attachment=attachment, content_type=content_type)


@shared_task(name='invoices.send_invoice_report')
def send_invoice_report():
    """ Sends aggregate accounting data as CSV """
    date = get_previous_month()
    subject = render_to_string('invoices/report_subject.txt', {
        'month': date.month,
        'year': date.year,
    }).strip()
    body = render_to_string('invoices/report_body.txt', {
        'month': date.month,
        'year': date.year,
    }).strip()
    filename = '3M%02d%dWaldur.txt' % (date.month, date.year)
    invoices = models.Invoice.objects.filter(year=date.year, month=date.month)

    # Report should include only organizations that had accounting running during the invoice period.
    if settings.WALDUR_CORE['ENABLE_ACCOUNTING_START_DATE']:
        invoices = invoices.filter(customer__accounting_start_date__lte=core_utils.month_end(date))

    # Report should not include customers with 0 invoice sum.
    invoices = [invoice for invoice in invoices if invoice.total > 0]
    text_message = format_invoice_csv(invoices)

    # Please note that email body could be empty if there are no valid invoices
    emails = [settings.WALDUR_INVOICES['INVOICE_REPORTING']['EMAIL']]
    logger.debug('About to send accounting report to {emails}'.format(emails=emails))
    core_utils.send_mail_with_attachment(
        subject=subject,
        body=body,
        to=emails,
        attachment=text_message,
        filename=filename
    )


def format_invoice_csv(invoices):
    if not isinstance(invoices, list):
        invoices = [invoices]

    csv_params = settings.WALDUR_INVOICES['INVOICE_REPORTING']['CSV_PARAMS']

    if settings.WALDUR_INVOICES['INVOICE_REPORTING'].get('USE_SAF'):
        fields = serializers.SAFReportSerializer.Meta.fields
        stream = StringIO()
        writer = DictWriter(stream, fieldnames=fields, **csv_params)
        writer.writeheader()

        for invoice in invoices:
            items = invoice.items
            items = utils.filter_invoice_items(items)
            serializer = serializers.SAFReportSerializer(items, many=True)
            writer.writerows(serializer.data)
        return stream.getvalue()

    fields = serializers.InvoiceItemReportSerializer.Meta.fields
    stream = StringIO()
    writer = DictWriter(stream, fieldnames=fields, **csv_params)
    writer.writeheader()

    for invoice in invoices:
        items = invoice.items
        items = utils.filter_invoice_items(items)
        serializer = serializers.InvoiceItemReportSerializer(items, many=True)
        writer.writerows(serializer.data)

    return stream.getvalue()


@shared_task(name='invoices.update_invoices_current_cost')
def update_invoices_current_cost():
    year = utils.get_current_year()
    month = utils.get_current_month()

    for invoice in models.Invoice.objects.filter(year=year, month=month):
        invoice.update_current_cost()


@shared_task
def create_invoice_pdf(serialized_invoice):
    invoice = core_utils.deserialize_instance(serialized_invoice)
    utils.create_invoice_pdf(invoice)


@shared_task
def create_pdf_for_all_invoices():
    for invoice in models.Invoice.objects.all():
        utils.create_invoice_pdf(invoice)


@shared_task
def create_pdf_for_new_invoices():
    date = timezone.now()
    for invoice in models.Invoice.objects.filter(year=date.year, month=date.month):
        utils.create_invoice_pdf(invoice)


@shared_task
def send_new_invoices_notification():
    date = timezone.now()

    for invoice in models.Invoice.objects.filter(year=date.year, month=date.month):
        send_invoice_notification.delay(invoice.uuid.hex)



"""
Django base settings for Waldur Core.
"""
from datetime import timedelta
import locale
# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os
import warnings

from celery.schedules import crontab

from waldur_core.core import WaldurExtension
from waldur_core.server.admin.settings import *  # noqa: F403

encoding = locale.getpreferredencoding()
if encoding.lower() != 'utf-8':
    raise Exception("""Your system's preferred encoding is `{}`, but Waldur requires `UTF-8`.
Fix it by setting the LC_* and LANG environment settings. Example:
LC_ALL=en_US.UTF-8 LANG=en_US.UTF-8
""".format(encoding))

ADMINS = ()

BASE_DIR = os.path.abspath(os.path.join(os.path.join(os.path.dirname(os.path.dirname(__file__)), '..'), '..'))

DEBUG = False

MEDIA_ROOT = '/media_root/'

MEDIA_URL = '/media/'

ALLOWED_HOSTS = []

# Application definition
INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.humanize',
    'django.contrib.staticfiles',

    'waldur_core.landing',
    'waldur_core.logging',
    'waldur_core.core',
    'waldur_core.monitoring',
    'waldur_core.quotas',
    'waldur_core.structure',
    'waldur_core.cost_tracking',
    'waldur_core.users',
    'waldur_core.media',

    'rest_framework',
    'rest_framework.authtoken',
    'rest_framework_swagger',
    'django_filters',

    'defender',
    'django_fsm',
    'reversion',
    'taggit',
    'jsoneditor',
)
INSTALLED_APPS += ADMIN_INSTALLED_APPS  # noqa: F405

MIDDLEWARE = (
    'waldur_core.server.middleware.cors_middleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'waldur_core.logging.middleware.CaptureEventContextMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'defender.middleware.FailedLoginMiddleware',
)

REST_FRAMEWORK = {
    'TEST_REQUEST_DEFAULT_FORMAT': 'json',
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'waldur_core.core.authentication.TokenAuthentication',
        'waldur_core.core.authentication.SessionAuthentication',
    ),
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.IsAuthenticated',
    ),
    'DEFAULT_FILTER_BACKENDS': ('django_filters.rest_framework.DjangoFilterBackend',),
    'DEFAULT_RENDERER_CLASSES': (
        'rest_framework.renderers.JSONRenderer',
        'waldur_core.core.renderers.BrowsableAPIRenderer',
    ),
    'DEFAULT_PAGINATION_CLASS': 'waldur_core.core.pagination.LinkHeaderPagination',
    'DEFAULT_SCHEMA_CLASS': 'rest_framework.schemas.coreapi.AutoSchema',
    'PAGE_SIZE': 10,
    'EXCEPTION_HANDLER': 'waldur_core.core.views.exception_handler',

    # Return native `Date` and `Time` objects in `serializer.data`
    'DATETIME_FORMAT': None,
    'DATE_FORMAT': None,
    'TIME_FORMAT': None,
    'ORDERING_PARAM': 'o'
}

AUTHENTICATION_BACKENDS = (
    'django.contrib.auth.backends.ModelBackend',
    'waldur_core.core.authentication.AuthenticationBackend',
)

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

ANONYMOUS_USER_ID = None

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'src', 'waldur_core', 'templates')],
        'OPTIONS': {
            'context_processors': (
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
            ),
            'loaders': (
                'django.template.loaders.filesystem.Loader',
                'django.template.loaders.app_directories.Loader',
            ) + ADMIN_TEMPLATE_LOADERS,  # noqa: F405
        },
    },
]

ROOT_URLCONF = 'waldur_core.server.urls'

AUTH_USER_MODEL = 'core.User'

# Session
# https://docs.djangoproject.com/en/2.2/ref/settings/#sessions
SESSION_COOKIE_AGE = 3600
SESSION_SAVE_EVERY_REQUEST = True

WSGI_APPLICATION = 'waldur_core.server.wsgi.application'

# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/
LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

LOCALE_PATHS = (
    os.path.join(BASE_DIR, 'src', 'waldur_core', 'locale'),
)

LANGUAGES = (
    ('en', 'English'),
    ('et', 'Estonian'),
)

USE_TZ = True

# Defender configuration
DEFENDER_REDIS_URL = 'redis://localhost:6379/0'

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/
STATIC_URL = '/static/'

# Celery
CELERY_BROKER_URL = 'redis://localhost'
CELERY_RESULT_BACKEND = 'redis://localhost'

CELERY_TASK_QUEUES = {
    'tasks': {'exchange': 'tasks'},
    'heavy': {'exchange': 'heavy'},
    'background': {'exchange': 'background'},
}
CELERY_TASK_DEFAULT_QUEUE = 'tasks'
CELERY_TASK_ROUTES = ('waldur_core.server.celery.PriorityRouter',)

CACHES = {
    'default': {
        'BACKEND': 'redis_cache.RedisCache',
        'LOCATION': 'redis://localhost',
        'OPTIONS': {
            'DB': 1,
            'PARSER_CLASS': 'redis.connection.HiredisParser',
            'CONNECTION_POOL_CLASS': 'redis.BlockingConnectionPool',
            'PICKLE_VERSION': -1,
        },
    },
}

# Regular tasks
CELERY_BEAT_SCHEDULE = {
    'pull-service-properties': {
        'task': 'waldur_core.structure.ServicePropertiesListPullTask',
        'schedule': timedelta(hours=24),
        'args': (),
    },
    'pull-service-resources': {
        'task': 'waldur_core.structure.ServiceResourcesListPullTask',
        'schedule': timedelta(hours=1),
        'args': (),
    },
    'pull-service-subresources': {
        'task': 'waldur_core.structure.ServiceSubResourcesListPullTask',
        'schedule': timedelta(hours=2),
        'args': (),
    },
    'check-expired-permissions': {
        'task': 'waldur_core.structure.check_expired_permissions',
        'schedule': timedelta(hours=24),
        'args': (),
    },
    'recalculate-price-estimates': {
        'task': 'waldur_core.cost_tracking.recalculate_estimate',
        # To avoid bugs and unexpected behavior - do not re-calculate estimates
        # right in the end of the month.
        'schedule': crontab(minute=10),
        'args': (),
    },
    'cancel-expired-invitations': {
        'task': 'waldur_core.users.cancel_expired_invitations',
        'schedule': timedelta(hours=24),
        'args': (),
    },
    'structure-set-erred-stuck-resources': {
        'task': 'waldur_core.structure.SetErredStuckResources',
        'schedule': timedelta(hours=1),
        'args': (),
    },
}

# Logging
# Send verified request on webhook processing
VERIFY_WEBHOOK_REQUESTS = True


# Extensions
WALDUR_CORE = {
    'EXTENSIONS_AUTOREGISTER': True,
    'TOKEN_KEY': 'x-auth-token',

    # wiki: http://docs.waldur.com/MasterMind+configuration
    'AUTHENTICATION_METHODS': [
        'LOCAL_SIGNIN',
    ],
    'INVITATIONS_ENABLED': True,
    'ALLOW_SIGNUP_WITHOUT_INVITATION': True,
    'VALIDATE_INVITATION_EMAIL': False,
    'TOKEN_LIFETIME': timedelta(hours=1),
    'INVITATION_LIFETIME': timedelta(weeks=1),
    'OWNERS_CAN_MANAGE_OWNERS': False,
    'OWNER_CAN_MANAGE_CUSTOMER': False,
    'BACKEND_FIELDS_EDITABLE': True,
    'INITIAL_CUSTOMER_AGREEMENT_NUMBER': 4000,
    'CREATE_DEFAULT_PROJECT_ON_ORGANIZATION_CREATION': False,
    'ONLY_STAFF_MANAGES_SERVICES': False,
    'COMPANY_TYPES': (
        'Ministry',
        'Private company',
        'Public company',
        'Government owned company',
    ),
    'NATIVE_NAME_ENABLED': False,
    'SITE_NAME': 'Waldur MasterMind',
    'SITE_ADDRESS': 'Default address',
    'SITE_EMAIL': 'Default email',
    'SITE_PHONE': 'Default phone',
    'SITE_LOGO': None,
    'CURRENCY_NAME': 'EUR',
    'LOGIN_COMPLETED_URL': 'https://example.com/#/login_completed/{token}/{method}/',
    'LOGIN_FAILED_URL': 'https://example.com/#/login_failed/',
    'LOGOUT_COMPLETED_URL': 'https://example.com/#/logout_completed/',
    'LOGOUT_FAILED_URL': 'https://example.com/#/logout_failed/',
    'NOTIFICATIONS_PROFILE_CHANGES': {'ENABLED': True, 'FIELDS': ('email', 'phone_number', 'job_title')},
    # 'COUNTRIES': ['EE', 'LV', 'LT'],
    'ENABLE_ACCOUNTING_START_DATE': False,
    'USE_ATOMIC_TRANSACTION': True,
    'NOTIFICATION_SUBJECT': 'Notifications from Waldur',
    'LOGGING_REPORT_DIRECTORY': '/var/log/waldur',
    'LOGGING_REPORT_INTERVAL': timedelta(days=7),
    'HTTP_CHUNK_SIZE': 50,
    'ONLY_STAFF_CAN_INVITE_USERS': False,
    'INVITATION_APPROVE_URL': 'https://example.com/#/invitation_approve/{token}/',
    'INVITATION_REJECT_URL': 'https://example.com/#/invitation_reject/{token}/',
    'INVITATION_MAX_AGE': None,
    'INVITATION_CREATE_MISSING_USER': False,
    'INVITATION_DISABLE_MULTIPLE_ROLES': False,
    'PROTECT_USER_DETAILS_FOR_REGISTRATION_METHODS': [],
    'ATTACHMENT_LINK_MAX_AGE': timedelta(hours=1),
    'EMAIL_CHANGE_URL': 'https://example.com/#/user_email_change/{code}/',
    'EMAIL_CHANGE_MAX_AGE': timedelta(days=1),
}

WALDUR_CORE_PUBLIC_SETTINGS = [
    'AUTHENTICATION_METHODS',
    'INVITATIONS_ENABLED',
    'ALLOW_SIGNUP_WITHOUT_INVITATION',
    'VALIDATE_INVITATION_EMAIL',
    'OWNER_CAN_MANAGE_CUSTOMER',
    'OWNERS_CAN_MANAGE_OWNERS',
    'COMPANY_TYPES',
    'NATIVE_NAME_ENABLED',
    'ONLY_STAFF_MANAGES_SERVICES',
    'PROTECT_USER_DETAILS_FOR_REGISTRATION_METHODS',
]

for ext in WaldurExtension.get_extensions():
    INSTALLED_APPS += (ext.django_app(),)

    for name, task in ext.celery_tasks().items():
        if name in CELERY_BEAT_SCHEDULE:
            warnings.warn(
                "Celery beat task %s from Waldur extension %s "
                "is overlapping with primary tasks definition" % (name, ext.django_app()))
        else:
            CELERY_BEAT_SCHEDULE[name] = task

    for key, val in ext.Settings.__dict__.items():
        if not key.startswith('_'):
            globals()[key] = val

    ext.update_settings(globals())


# Swagger
SWAGGER_SETTINGS = {
    # USE_SESSION_AUTH parameter should be equal to DEBUG parameter.
    # If it is True, LOGIN_URL and LOGOUT_URL must be specified.
    'USE_SESSION_AUTH': False,
    'APIS_SORTER': 'alpha',
    'JSON_EDITOR': True,
    'SECURITY_DEFINITIONS': {
        'api_key': {
            'type': 'apiKey',
            'name': 'Authorization',
            'in': 'header',
        },
    },
}

IPSTACK_ACCESS_KEY = ''
